# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""
Support for building mpi4py with distutils/setuptools.
"""

# ruff: noqa: E402
# ruff: noqa: FURB101, FURB103
# ruff: noqa: PTH106, PTH107, PTH109
# ruff: noqa: PTH110, PTH118, PTH119
# ruff: noqa: PTH120, PTH122, PTH123
# ruff: noqa: PTH207

# -----------------------------------------------------------------------------

import contextlib
import glob
import os
import platform
import re
import shlex
import sys
import warnings
from distutils import log, sysconfig
from distutils.file_util import copy_file
from distutils.util import convert_path

# Fix missing variables PyPy's  distutils.sysconfig
if hasattr(sys, "pypy_version_info"):
    config_vars = sysconfig.get_config_vars()
    for name in ("prefix", "exec_prefix"):
        if name not in config_vars:
            config_vars[name] = os.path.normpath(getattr(sys, name))
    if sys.platform == "darwin" and "LDSHARED" in config_vars:
        ldshared = shlex.split(config_vars["LDSHARED"])
        while "-shared" in ldshared:
            ldshared[ldshared.index("-shared")] = "-bundle"
        if "-undefined" not in ldshared:
            ldshared.extend("-undefined dynamic_lookup".split())
        config_vars["LDSHARED"] = " ".join(ldshared)

# Workaround distutils.cygwinccompiler.get_versions()
# failing when the compiler path contains spaces
from distutils import cygwinccompiler as cygcc

if hasattr(cygcc, "get_versions"):
    cygcc_get_versions = cygcc.get_versions

    def get_versions():
        import distutils.spawn

        find_executable_orig = distutils.spawn.find_executable

        def find_executable(exe):
            exe = find_executable_orig(exe)
            if exe and " " in exe:
                exe = f'"{exe}"'
            return exe

        distutils.spawn.find_executable = find_executable
        versions = cygcc_get_versions()
        distutils.spawn.find_executable = find_executable_orig
        return versions

    cygcc.get_versions = get_versions

# Workaround distutils.ccompiler.CCompiler._fix_lib_args
from distutils.ccompiler import CCompiler

cc_fix_compile_args_orig = getattr(CCompiler, "_fix_compile_args", None)
cc_fix_lib_args_orig = getattr(CCompiler, "_fix_lib_args", None)


def cc_fix_compile_args(self, out_dir, macros, inc_dirs):
    macros = macros or []
    inc_dirs = inc_dirs or []
    return cc_fix_compile_args_orig(self, out_dir, macros, inc_dirs)


def cc_fix_lib_args(self, libs, lib_dirs, rt_lib_dirs):
    libs = libs or []
    lib_dirs = lib_dirs or []
    rt_lib_dirs = rt_lib_dirs or []
    return cc_fix_lib_args_orig(self, libs, lib_dirs, rt_lib_dirs)


CCompiler._fix_compile_args = cc_fix_compile_args
CCompiler._fix_lib_args = cc_fix_lib_args


def _fix_env(cmd, i):
    while os.path.basename(cmd[i]) == "env":
        i = i + 1
        while "=" in cmd[i]:
            i = i + 1
    return i


def _fix_xcrun(cmd, i):
    if os.path.basename(cmd[i]) == "xcrun":
        del cmd[i]
        while True:
            if cmd[i] == "-sdk":
                del cmd[i : i + 2]
                continue
            if cmd[i] == "-log":
                del cmd[i]
                continue
            break
    return i


def fix_compiler_cmd(cc, mpicc):
    if not mpicc:
        return
    i = 0
    i = _fix_env(cc, i)
    i = _fix_xcrun(cc, i)
    while os.path.basename(cc[i]) == "ccache":
        i = i + 1
    cc[i : i + 1] = shlex.split(mpicc)


def fix_linker_cmd(ld, mpild):
    if not mpild:
        return
    i = 0
    if (
        sys.platform.startswith("aix")
        and os.path.basename(ld[i]) == "ld_so_aix"
    ):
        i = 1
    i = _fix_env(ld, i)
    i = _fix_xcrun(ld, i)
    while os.path.basename(ld[i]) == "ccache":
        del ld[i]
    ld[i : i + 1] = shlex.split(mpild)


def customize_compiler(
    compiler,
    lang=None,
    mpicc=None,
    mpicxx=None,
    mpild=None,
):
    sysconfig.customize_compiler(compiler)
    if compiler.compiler_type == "unix":
        ld = compiler.linker_exe
        for envvar in ("LDFLAGS", "CFLAGS", "CPPFLAGS"):
            if envvar in os.environ:
                ld += shlex.split(os.environ[envvar])
    if os.environ.get("SOURCE_DATE_EPOCH") is not None:
        # Linker tweaks for reproducible build
        if sys.platform == "darwin":
            os.environ["ZERO_AR_DATE"] = "YES"
        if compiler.compiler_type == "msvc":
            if not compiler.initialized:
                compiler.initialize()
            for flags in compiler._ldflags.values():
                flags.append("/BREPRO")
    if compiler.compiler_type == "unix":
        # Compiler command overriding
        if mpicc:
            fix_compiler_cmd(compiler.compiler, mpicc)
            if lang in ("c", None):
                fix_compiler_cmd(compiler.compiler_so, mpicc)
        if mpicxx:
            fix_compiler_cmd(compiler.compiler_cxx, mpicxx)
            if lang == "c++":
                fix_compiler_cmd(compiler.compiler_so, mpicxx)
        if mpild:
            for ld in [compiler.linker_so, compiler.linker_exe]:
                fix_linker_cmd(ld, mpild)
    if compiler.compiler_type == "cygwin":
        compiler.set_executables(
            preprocessor="gcc -mcygwin -E",
        )
    if compiler.compiler_type == "mingw32":
        compiler.set_executables(
            preprocessor="gcc -mno-cygwin -E",
        )
    if compiler.compiler_type in ("unix", "cygwin", "mingw32"):
        badcxxflags = ["-Wimplicit", "-Wstrict-prototypes"]
        for flag in badcxxflags:
            while flag in compiler.compiler_cxx:
                compiler.compiler_cxx.remove(flag)
            if lang == "c++":
                while flag in compiler.compiler_so:
                    compiler.compiler_so.remove(flag)
    if compiler.compiler_type == "mingw32":
        # Remove msvcrXX.dll
        del compiler.dll_libraries[:]
        # https://bugs.python.org/issue12641
        if compiler.gcc_version >= "4.4":
            for attr in (
                "preprocessor",
                "compiler",
                "compiler_cxx",
                "compiler_so",
                "linker_so",
                "linker_exe",
            ):
                with contextlib.suppress(Exception):
                    getattr(compiler, attr).remove("-mno-cygwin")
        # Add required define and compiler flags for AMD64
        if platform.architecture(None)[0] == "64bit":
            for attr in (
                "preprocessor",
                "compiler",
                "compiler_cxx",
                "compiler_so",
                "linker_so",
                "linker_exe",
            ):
                getattr(compiler, attr).insert(1, "-DMS_WIN64")
                getattr(compiler, attr).insert(1, "-m64")


# -----------------------------------------------------------------------------

from mpiconfig import Config


def configuration(command_obj, verbose=True):
    config = Config(log)
    config.setup(command_obj)
    if verbose:
        if config.section and config.filename:
            config.log.info(
                "MPI configuration: [%s] from '%s'",
                config.section,
                ",".join(config.filename),
            )
            config.info()
    return config


def configure_compiler(compiler, config, lang=None):
    #
    mpicc = config.get("mpicc")
    mpicxx = config.get("mpicxx")
    mpild = config.get("mpild")
    if not mpild and (mpicc or mpicxx):
        if lang == "c":
            mpild = mpicc
        if lang == "c++":
            mpild = mpicxx
        if not mpild:
            mpild = mpicc or mpicxx
    #
    customize_compiler(
        compiler,
        lang,
        mpicc=mpicc,
        mpicxx=mpicxx,
        mpild=mpild,
    )
    #
    for k, v in config.get("define_macros", []):
        compiler.define_macro(k, v)
    for v in config.get("undef_macros", []):
        compiler.undefine_macro(v)
    for v in config.get("include_dirs", []):
        compiler.add_include_dir(v)
    for v in config.get("libraries", []):
        compiler.add_library(v)
    for v in config.get("library_dirs", []):
        compiler.add_library_dir(v)
    for v in config.get("runtime_library_dirs", []):
        compiler.add_runtime_library_dir(v)
    for v in config.get("extra_objects", []):
        compiler.add_link_object(v)
    if compiler.compiler_type in (
        "unix",
        "intel",
        "cygwin",
        "mingw32",
    ):
        cc_args = config.get("extra_compile_args", [])
        ld_args = config.get("extra_link_args", [])
        compiler.compiler += cc_args
        compiler.compiler_so += cc_args
        compiler.compiler_cxx += cc_args
        compiler.linker_so += ld_args
        compiler.linker_exe += ld_args
    return compiler


# -----------------------------------------------------------------------------

try:
    from mpiapigen import Generator
except ImportError:

    class Generator:
        def parse_file(self, *args):
            raise NotImplementedError("You forgot to grab 'mpiapigen.py'")


@contextlib.contextmanager
def capture_stderr(filename=os.devnull):
    stream = sys.stderr
    file_obj = None
    fno_save = None
    try:
        file_obj = open(filename, "w")
        fno_save = os.dup(stream.fileno())
        os.dup2(file_obj.fileno(), stream.fileno())
        yield
    finally:
        if file_obj is not None:
            file_obj.close()
        if fno_save is not None:
            os.dup2(fno_save, stream.fileno())


class ConfigureMPI:
    SRCDIR = "src"
    SOURCES = [os.path.join("mpi4py", "libmpi.pxd")]
    DESTDIR = os.path.join("src", "lib-mpi")
    CONFIG_H = "pympiconf.h"
    MISSING_H = "missing.h"

    CONFIGTEST_H = """\
/* _configtest.h */

#if !defined(MPIAPI)
#  define MPIAPI
#endif

"""

    def __init__(self, config_cmd):
        self.generator = Generator()
        for filename in self.SOURCES:
            fullname = os.path.join(self.SRCDIR, filename)
            self.generator.parse_file(fullname)
        self.config_cmd = config_cmd

    def run(self):
        results = []
        with open("_configtest.h", "w") as f:
            f.write(self.CONFIGTEST_H)
        for node in self.generator:
            name = node.name
            testcode = node.config()
            confcode = node.missing(guard=False)
            log.info("checking for '%s'...", name)
            ok = self.run_test(testcode)
            if not ok:
                log.info("**** failed check for '%s'", name)
                with open("_configtest.h", "a") as f:
                    f.write(confcode)
            results.append((name, ok))
        try:
            os.remove("_configtest.h")
        except OSError:
            pass
        return results

    def gen_test(self, code):
        body = [
            '#include "_configtest.h"',
            "int main(int argc, char **argv) {",
            "\n".join(["  " + line for line in code.split("\n")]),
            "  (void)argc; (void)argv;",
            "  return 0;",
            "}",
        ]
        body = "\n".join(body) + "\n"
        return body

    def run_test(self, code, lang="c"):
        level = log.set_threshold(log.WARN)
        log.set_threshold(level)
        if not self.config_cmd.noisy:
            level = log.set_threshold(log.WARN)
        try:
            body = self.gen_test(code)
            headers = ["stdlib.h", "mpi.h"]
            ok = self.config_cmd.try_link(body, headers=headers, lang=lang)
            return ok
        finally:
            log.set_threshold(level)

    def dump(self, results):
        destdir = self.DESTDIR
        config_h = os.path.join(destdir, self.CONFIG_H)
        missing_h = os.path.join(destdir, self.MISSING_H)
        log.info("writing '%s'", config_h)
        self.generator.dump_config_h(config_h, results)
        log.info("writing '%s'", missing_h)
        self.generator.dump_missing_h(missing_h, None)


# -----------------------------------------------------------------------------

cmd_mpi_opts = [
    (
        "mpild=",
        None,
        "MPI linker command, "
        "overridden by environment variable 'MPILD' "
        "(defaults to 'mpicc' or 'mpicxx' if any is available)",
    ),
    (
        "mpicxx=",
        None,
        "MPI C++ compiler command, "
        "overridden by environment variable 'MPICXX' "
        "(defaults to 'mpicxx', 'mpiCC', or 'mpic++' if any is available)",
    ),
    (
        "mpicc=",
        None,
        "MPI C compiler command, "
        "overridden by environment variables 'MPICC' "
        "(defaults to 'mpicc' if available)",
    ),
    (
        "mpi=",
        None,
        "specify a ini-style configuration file and section "
        "(e.g. --mpi=filename or --mpi=filename:section), "
        "to look for MPI includes/libraries, "
        "overridden by environment variable 'MPICFG' "
        "(defaults to configuration file 'mpi.cfg' and section 'mpi')",
    ),
    (
        "configure",
        None,
        "exhaustive test for checking missing MPI constants/types/functions",
    ),
]


def cmd_get_mpi_options(cmd_opts):
    optlist = []
    for option, _, _ in cmd_opts:
        if option[-1] == "=":
            option = option[:-1]
        option = option.replace("-", "_")
        optlist.append(option)
    return optlist


def cmd_initialize_mpi_options(cmd):
    mpiopts = cmd_get_mpi_options(cmd_mpi_opts)
    for op in mpiopts:
        setattr(cmd, op, None)


def cmd_set_undefined_mpi_options(cmd, basecmd):
    mpiopts = cmd_get_mpi_options(cmd_mpi_opts)
    optlist = tuple(zip(mpiopts, mpiopts))
    cmd.set_undefined_options(basecmd, *optlist)


# -----------------------------------------------------------------------------

try:
    import setuptools
except ImportError:
    setuptools = None


def import_command(cmd):
    from importlib import import_module

    if setuptools:
        try:
            return import_module("setuptools.command." + cmd)
        except ImportError:
            pass
    return import_module("distutils.command." + cmd)


if setuptools:
    from setuptools import (
        Command,
        Distribution as cls_Distribution,
        Extension as cls_Extension,
    )
else:
    from distutils.core import (
        Command,
        Distribution as cls_Distribution,
        Extension as cls_Extension,
    )

cmd_config = import_command("config")
cmd_build = import_command("build")
cmd_install = import_command("install")
cmd_clean = import_command("clean")

cmd_build_ext = import_command("build_ext")
cmd_install_lib = import_command("install_lib")
cmd_install_data = import_command("install_data")

from distutils.errors import (
    CCompilerError,
    DistutilsError,
    DistutilsPlatformError,
    DistutilsSetupError,
)

try:
    from packaging.version import Version
except ImportError:
    try:
        from setuptools.extern.packaging.version import Version
    except ImportError:
        from distutils.version import StrictVersion as Version
try:
    from setuptools.modified import newer_group
except ImportError:
    try:
        from setuptools.dep_util import newer_group
    except ImportError:
        from distutils.dep_util import newer_group

# -----------------------------------------------------------------------------

# Distribution class supporting a 'executables' keyword


class Distribution(cls_Distribution):
    def __init__(self, attrs=None):
        # support for pkg data
        self.package_data = {}
        # PEP 314
        self.provides = None
        self.requires = None
        self.obsoletes = None
        # supports 'executables' keyword
        self.executables = None
        cls_Distribution.__init__(self, attrs)

    def has_executables(self):
        return self.executables and len(self.executables) > 0

    def is_pure(self):
        return cls_Distribution.is_pure(self) and not self.has_executables()


# Extension class


class Extension(cls_Extension):
    def __init__(self, **kw):
        optional = kw.pop("optional", None)
        configure = kw.pop("configure", None)
        cls_Extension.__init__(self, **kw)
        self.optional = optional
        self.configure = configure


# Library class


class Library(Extension):
    def __init__(self, **kw):
        kind = kw.pop("kind", "static")
        package = kw.pop("package", None)
        dest_dir = kw.pop("dest_dir", None)
        Extension.__init__(self, **kw)
        self.kind = kind
        self.package = package
        self.dest_dir = dest_dir


# Executable class


class Executable(Extension):
    def __init__(self, **kw):
        package = kw.pop("package", None)
        dest_dir = kw.pop("dest_dir", None)
        Extension.__init__(self, **kw)
        self.package = package
        self.dest_dir = dest_dir


# setup function


def setup(**attrs):
    if setuptools:
        from setuptools import setup as fcn_setup
    else:
        from distutils.core import setup as fcn_setup
    if "distclass" not in attrs:
        attrs["distclass"] = Distribution
    if "cmdclass" not in attrs:
        attrs["cmdclass"] = {}
    cmdclass = attrs["cmdclass"]
    for cmd in (
        config,
        build,
        install,
        clean,
        build_src,
        build_ext,
        build_exe,
        install_lib,
        install_data,
        install_exe,
    ):
        if cmd.__name__ not in cmdclass:
            cmdclass[cmd.__name__] = cmd
    return fcn_setup(**attrs)


# --------------------------------------------------------------------


def with_coverage():
    return bool(os.environ.get("MPI4PY_COVERAGE_PLUGIN"))


# --------------------------------------------------------------------

# Cython


def cython_req():
    confdir = os.path.dirname(__file__)
    basename = "requirements-build-cython.txt"
    with open(os.path.join(confdir, basename)) as f:
        m = re.search(r"cython\s*>?=+\s*(.*)", f.read().strip())
    cython_version = m.groups()[0]
    return cython_version


def cython_chk(VERSION, verbose=True):
    #
    def warn(message):
        if not verbose:
            return
        ruler, ws, nl = "*" * 80, " ", "\n"
        pyexe = sys.executable
        advise = f"$ {pyexe} -m pip install --upgrade cython"

        def printer(*s):
            print(*s, file=sys.stderr)

        printer(ruler, nl)
        printer(ws, message, nl)
        printer(ws, ws, advise, nl)
        printer(ruler)

    #
    try:
        import Cython
    except ImportError:
        warn("You need Cython to generate C source files.")
        return False
    #
    CYTHON_VERSION = Cython.__version__
    m = re.match(r"(\d+\.\d+(?:\.\d+)?).*", CYTHON_VERSION)
    if not m:
        warn(f"Cannot parse Cython version string {CYTHON_VERSION!r}")
        return False
    REQUIRED = Version(VERSION)
    PROVIDED = Version(m.groups()[0])
    if PROVIDED < REQUIRED:
        warn(f"You need Cython >= {VERSION} (you have {CYTHON_VERSION})")
        return False
    #
    if verbose:
        log.info("using Cython %s", CYTHON_VERSION)
    return True


def cython_run(
    source,
    target=None,
    depends=(),
    includes=(),
    workdir=None,
    force=False,
    VERSION="0.0",
):
    if target is None:
        target = os.path.splitext(source)[0] + ".c"
    cwd = os.getcwd()
    try:
        if workdir:
            os.chdir(workdir)
        alldeps = [source]
        for dep in depends:
            alldeps += glob.glob(dep)
        if not (force or newer_group(alldeps, target, "newer")):
            log.debug("skipping '%s' -> '%s' (up-to-date)", source, target)
            return
    finally:
        os.chdir(cwd)
    require = f"Cython >= {VERSION}"
    if not cython_chk(VERSION, verbose=False) and setuptools:
        if sys.modules.get("Cython"):
            removed = getattr(sys.modules["Cython"], "__version__", "")
            log.info("removing Cython %s from sys.modules", removed)
            pkgname = re.compile(r"cython(\.|$)", re.IGNORECASE)
            for modname in list(sys.modules.keys()):
                if pkgname.match(modname):
                    del sys.modules[modname]
        try:
            install_setup_requires = setuptools._install_setup_requires
            with warnings.catch_warnings():
                category = setuptools.SetuptoolsDeprecationWarning
                warnings.simplefilter("ignore", category)
                log.info("fetching build requirement '%s'", require)
                install_setup_requires({"setup_requires": [require]})
        except Exception:
            log.info("failed to fetch build requirement '%s'", require)
    if not cython_chk(VERSION):
        raise DistutilsError(f"missing build requirement {require!r}")
    #
    log.info("cythonizing '%s' -> '%s'", source, target)
    from cythonize import cythonize

    args = []
    if with_coverage():
        args += ["-X", "linetrace=True"]
    if includes:
        args += [f"-I{incdir}" for incdir in includes]
    if workdir:
        args += ["--working", workdir]
    args += [source]
    if target:
        args += ["--output-file", target]
    err = cythonize(args)
    if err:
        raise DistutilsError(f"Cython failure: {source!r} -> {target!r}")


# -----------------------------------------------------------------------------

# A minimalistic MPI program :-)

ConfigTest = """\
int main(int argc, char **argv)
{
  int ierr;
  (void)argc; (void)argv;
  ierr = MPI_Init(&argc, &argv);
  if (ierr) return -1;
  ierr = MPI_Finalize();
  if (ierr) return -1;
  return 0;
}
"""


class config(cmd_config.config):
    user_options = cmd_config.config.user_options + cmd_mpi_opts

    def initialize_options(self):
        cmd_config.config.initialize_options(self)
        cmd_initialize_mpi_options(self)
        self.noisy = 0

    def finalize_options(self):
        cmd_config.config.finalize_options(self)
        if not self.noisy:
            self.dump_source = 0

    def _clean(self, *a, **kw):
        if sys.platform.startswith("win"):
            for fn in ("_configtest.exe.manifest",):
                if os.path.exists(fn):
                    self.temp_files.append(fn)
        cmd_config.config._clean(self, *a, **kw)

    def check_header(
        self,
        header,
        headers=None,
        include_dirs=None,
    ):
        if headers is None:
            headers = []
        log.info("checking for header '%s' ...", header)
        body = "int main(int n, char**v) { (void)n; (void)v; return 0; }"
        ok = self.try_compile(body, [*headers, header], include_dirs)
        log.info("success!" if ok else "failure.")
        return ok

    def check_macro(
        self,
        macro,
        headers=None,
        include_dirs=None,
    ):
        log.info("checking for macro '%s' ...", macro)
        body = [
            f"#ifndef {macro}",
            f"#error macro '{macro}' not defined",
            r"#endif",
            r"int main(int n, char**v) { (void)n; (void)v; return 0; }",
        ]
        body = "\n".join(body) + "\n"
        ok = self.try_compile(body, headers, include_dirs)
        return ok

    def check_library(
        self,
        library,
        library_dirs=None,
        headers=None,
        include_dirs=None,
        other_libraries=(),
        lang="c",
    ):
        if sys.platform == "darwin":
            self.compiler.linker_exe.append("-flat_namespace")
            self.compiler.linker_exe.append("-undefined")
            self.compiler.linker_exe.append("suppress")
        log.info("checking for library '%s' ...", library)
        body = "int main(int n, char**v) { (void)n; (void)v; return 0; }"
        ok = self.try_link(
            body,
            headers,
            include_dirs,
            [library, *other_libraries],
            library_dirs,
            lang=lang,
        )
        if sys.platform == "darwin":
            self.compiler.linker_exe.remove("-flat_namespace")
            self.compiler.linker_exe.remove("-undefined")
            self.compiler.linker_exe.remove("suppress")
        return ok

    def check_function(
        self,
        function,
        headers=None,
        include_dirs=None,
        libraries=None,
        library_dirs=None,
        decl=0,
        call=0,
        lang="c",
    ):
        log.info("checking for function '%s' ...", function)
        body = []
        if decl:
            if call:
                proto = f"int {function} (void);"
            else:
                proto = f"int {function};"
            if lang == "c":
                proto = "\n".join([
                    "#ifdef __cplusplus",
                    'extern "C"',
                    "#endif",
                    proto,
                ])
            body.append(proto)
        body.append(r"int main (int n, char**v) {")
        if call:
            body.append(f"  (void){function}();")
        else:
            body.append(f"  {function};")
        body.extend((r"  (void)n; (void)v;", r"  return 0;", r"}"))
        body = "\n".join(body) + "\n"
        ok = self.try_link(
            body,
            headers,
            include_dirs,
            libraries,
            library_dirs,
            lang=lang,
        )
        return ok

    def check_symbol(
        self,
        symbol,
        type="int",  # noqa: A002
        headers=None,
        include_dirs=None,
        libraries=None,
        library_dirs=None,
        decl=0,
        lang="c",
    ):
        log.info("checking for symbol '%s' ...", symbol)
        body = []
        if decl:
            body.append(f"{type} {symbol};")
        body.extend((
            r"int main (int n, char**v) {",
            f"  {type} s; s = {symbol}; (void)s;",
            r"  (void)n; (void)v;",
            r"  return 0;",
            r"}",
        ))
        body = "\n".join(body) + "\n"
        ok = self.try_link(
            body,
            headers,
            include_dirs,
            libraries,
            library_dirs,
            lang=lang,
        )
        return ok

    def check_function_call(
        self,
        function,
        args="",
        headers=None,
        include_dirs=None,
        libraries=None,
        library_dirs=None,
        lang="c",
    ):
        log.info("checking for function '%s' ...", function)
        body = []
        body.extend((
            r"int main (int n, char**v) {",
            f"  (void){function}({args});",
            r"  (void)n; (void)v;",
            r"  return 0;",
            r"}",
        ))
        body = "\n".join(body) + "\n"
        ok = self.try_link(
            body,
            headers,
            include_dirs,
            libraries,
            library_dirs,
            lang=lang,
        )
        return ok

    def run(self):
        config = configuration(self, verbose=True)
        # test MPI C compiler
        self.compiler = getattr(self.compiler, "compiler_type", self.compiler)
        self._check_compiler()
        configure_compiler(self.compiler, config, lang="c")
        self.try_link(ConfigTest, headers=["mpi.h"], lang="c")
        # test MPI C++ compiler
        self.compiler = getattr(self.compiler, "compiler_type", self.compiler)
        self._check_compiler()
        configure_compiler(self.compiler, config, lang="c++")
        self.try_link(ConfigTest, headers=["mpi.h"], lang="c++")


def configure_dl(ext, config_cmd):
    log.info("checking for dlopen() availability ...")
    dlfcn = config_cmd.check_header("dlfcn.h")
    libdl = config_cmd.check_library("dl")
    libs = ["dl"] if libdl else None
    dlopen = config_cmd.check_function(
        "dlopen", libraries=libs, decl=1, call=1
    )
    if dlfcn:
        ext.define_macros += [("HAVE_DLFCN_H", 1)]
    if dlopen:
        ext.define_macros += [("HAVE_DLOPEN", 1)]


def configure_mpi(ext, config_cmd):
    from textwrap import dedent

    headers = ["stdlib.h", "mpi.h"]
    #
    log.info("checking for MPI compile and link ...")
    ConfigTest = dedent("""\
    int main(int argc, char **argv)
    {
      (void)MPI_Init(&argc, &argv);
      (void)MPI_Finalize();
      return 0;
    }
    """)
    errmsg = [
        "Cannot {} MPI programs. Check your configuration!!!",
        "Installing mpi4py requires a working MPI implementation.",
    ]
    if sys.platform == "linux":
        errmsg += [
            "If you are running on a supercomputer or cluster, check with",
            "the system administrator or refer to the system user guide.",
            "Otherwise, if you are running on a laptop or desktop computer,",
            "your may be missing the MPICH or Open MPI development package:",
            "* On Fedora/RHEL systems, run:",
            "  $ sudo dnf install mpich-devel     # for MPICH",
            "  $ sudo dnf install openmpi-devel   # for Open MPI",
            "* On Debian/Ubuntu systems, run:",
            "  $ sudo apt install libmpich-dev    # for MPICH",
            "  $ sudo apt install libopenmpi-dev  # for Open MPI",
        ]
    if sys.platform == "darwin":
        errmsg += [
            "Install MPICH or Open MPI with Homebrew or MacPorts:"
            "  $ brew install mpich|openmpi  # Homebrew",
            "  $ port install mpich|openmpi  # MacPorts",
        ]
    if sys.platform == "win32":
        errmsg += ["Please install *Intel MPI* or *Microsoft MPI*."]
    ok = config_cmd.try_compile(ConfigTest, headers=headers)
    if not ok:
        message = "\n".join(errmsg).format("compile")
        raise DistutilsPlatformError(message)
    ok = config_cmd.try_link(ConfigTest, headers=headers)
    if not ok:
        message = errmsg[0].format("link")
        raise DistutilsPlatformError(message)
    #
    impls = ("OPEN_MPI", "MSMPI_VER")
    tests = [f"defined({macro})" for macro in impls]
    tests += ["(defined(MPICH_NAME)&&(MPICH_NAME>=3))"]
    tests += ["(defined(MPICH_NAME)&&(MPICH_NAME==2))"]
    tests = "||".join(tests)
    ConfigTestAPI = dedent(f"""\
    #if !({tests})
    #error "Unknown MPI implementation"
    #endif
    """)
    ConfigTestABI = dedent("""\
    #if !(defined(MPI_ABI_VERSION)&&(MPI_ABI_VERSION>=1))
    #error "MPI ABI not supported"
    #endif
    """)
    with capture_stderr():
        log.info("checking for MPI ABI support ...")
        mpiabi = config_cmd.try_compile(ConfigTestABI, headers=headers)
    config = os.environ.get("MPI4PY_BUILD_CONFIGURE") or None
    config = getattr(config_cmd, "configure", None) or config
    if not mpiabi and not config:
        with capture_stderr():
            ok = config_cmd.try_compile(ConfigTestAPI, headers=headers)
            config = not ok
    if config:
        guard = "HAVE_PYMPICONF_H"
        with capture_stderr():
            ok = config_cmd.check_macro(guard)
        config = not ok
        if config:
            log.info("checking for missing MPI functions/symbols ...")
            configure = ConfigureMPI(config_cmd)
            with capture_stderr():
                results = configure.run()
            configure.dump(results)
            ext.define_macros += [(guard, 1)]
    elif not mpiabi:
        log.info("checking for missing MPI functions/symbols ...")
        for function, arglist in (
            ("MPI_Type_create_f90_integer", "0,(MPI_Datatype*)0"),
            ("MPI_Type_create_f90_real", "0,0,(MPI_Datatype*)0"),
            ("MPI_Type_create_f90_complex", "0,0,(MPI_Datatype*)0"),
            ("MPI_Status_c2f", "(MPI_Status*)0,(MPI_Fint*)0"),
            ("MPI_Status_f2c", "(MPI_Fint*)0,(MPI_Status*)0"),
        ):
            with capture_stderr():
                ok = config_cmd.check_function_call(
                    function, arglist, headers=headers
                )
            if not ok:
                macro = "PyMPI_MISSING_" + function
                ext.define_macros += [(macro, 1)]
    #
    if not mpiabi and os.name == "posix":
        configure_dl(ext, config_cmd)


def configure_pyexe(exe, _config_cmd):
    if sys.platform.startswith("win"):
        return
    if sys.platform == "darwin" and (
        "Anaconda" in sys.version or "Continuum Analytics" in sys.version
    ):
        py_version = sysconfig.get_python_version()
        py_abiflags = getattr(sys, "abiflags", "")
        exe.libraries += ["python" + py_version + py_abiflags]
        return
    #
    cfg_vars = sysconfig.get_config_vars()
    py_version = sysconfig.get_python_version()
    py_abiflags = getattr(sys, "abiflags", "")
    libraries = ["python" + py_version + py_abiflags]
    if hasattr(sys, "pypy_version_info"):
        py_tag = py_version[0].replace("2", "")
        libraries = [f"pypy{py_tag}-c"]
    if sys.platform == "darwin":
        fwkdir = cfg_vars.get("PYTHONFRAMEWORKDIR")
        if (
            fwkdir
            and fwkdir != "no-framework"
            and fwkdir in cfg_vars.get("LINKFORSHARED", "")
        ):
            del libraries[:]
    #
    py_enable_shared = cfg_vars.get("Py_ENABLE_SHARED")
    libdir = shlex.split(cfg_vars.get("LIBDIR", ""))
    libpl = shlex.split(cfg_vars.get("LIBPL", ""))
    library_dirs = []
    runtime_dirs = []
    if py_enable_shared:
        library_dirs += libdir
        if sys.exec_prefix != "/usr":
            runtime_dirs += libdir
    else:
        library_dirs += libdir
        library_dirs += libpl
    #
    link_args = []
    for var in ("LIBS", "MODLIBS", "SYSLIBS", "LDLAST"):
        link_args += shlex.split(cfg_vars.get(var, ""))
    #
    exe.libraries += libraries
    exe.library_dirs += library_dirs
    exe.runtime_library_dirs += runtime_dirs
    exe.extra_link_args += link_args


class build(cmd_build.build):
    user_options = cmd_build.build.user_options + cmd_mpi_opts
    boolean_options = cmd_build.build.boolean_options

    user_options += [
        (
            "inplace",
            "i",
            "ignore build-lib and put compiled extensions into the source "
            "directory alongside your pure Python modules",
        )
    ]
    boolean_options += ["inplace"]

    def initialize_options(self):
        cmd_build.build.initialize_options(self)
        cmd_initialize_mpi_options(self)
        self.inplace = None

    def finalize_options(self):
        cmd_build.build.finalize_options(self)
        config_cmd = self.get_finalized_command("config")
        if isinstance(config_cmd, config):
            cmd_set_undefined_mpi_options(self, "config")
        if self.inplace is None:
            self.inplace = False

    def has_executables(self):
        return self.distribution.has_executables()

    sub_commands = [
        ("build_src", lambda *_: True),
        *cmd_build.build.sub_commands,
        ("build_exe", has_executables),
    ]

    # NOTE: disable build_exe subcommand !!!
    del sub_commands[-1]


class build_src(Command):
    description = "build C sources from Cython files"

    user_options = [
        ("force", "f", "forcibly build everything (ignore file timestamps)"),
    ]

    boolean_options = ["force"]

    def initialize_options(self):
        self.force = False

    def finalize_options(self):
        self.set_undefined_options(
            "build",
            ("force", "force"),
        )

    def run(self):
        sources = getattr(self, "sources", [])
        require = cython_req()
        for source in sources:
            cython_run(
                **source,
                force=self.force,
                VERSION=require,
            )


class build_ext(cmd_build_ext.build_ext):
    user_options = cmd_build_ext.build_ext.user_options + cmd_mpi_opts

    def initialize_options(self):
        cmd_build_ext.build_ext.initialize_options(self)
        cmd_initialize_mpi_options(self)
        self.inplace = None

    def finalize_options(self):
        self.set_undefined_options("build", ("inplace", "inplace"))
        cmd_build_ext.build_ext.finalize_options(self)
        build_cmd = self.get_finalized_command("build")
        if isinstance(build_cmd, build):
            cmd_set_undefined_mpi_options(self, "build")

    def run(self):
        self.build_sources()
        cmd_build_ext.build_ext.run(self)

    def build_sources(self):
        if self.get_command_name() == "build_ext":
            if "build_src" in self.distribution.cmdclass:
                self.run_command("build_src")

    def build_extensions(self):
        # First, sanity-check the 'extensions' list
        self.check_extensions_list(self.extensions)
        # parse configuration file and configure compiler
        self.config = configuration(self, verbose=True)
        configure_compiler(self.compiler, self.config)
        # build extensions
        for ext in self.extensions:
            try:
                self.build_extension(ext)
            except (DistutilsError, CCompilerError):
                if not ext.optional:
                    raise
                e = sys.exc_info()[1]
                self.warn(f"{e}")
                exe = isinstance(ext, Executable)
                knd = "executable" if exe else "extension"
                self.warn(f'building optional {knd} "{ext.name}" failed')

    def config_extension(self, ext):
        configure = getattr(ext, "configure", None)
        if configure:
            config_cmd = self.get_finalized_command("config")
            config_cmd.compiler = self.compiler  # fix compiler
            config_cmd.configure = self.configure
            configure(ext, config_cmd)
        if with_coverage():
            ext.define_macros += [
                ("CYTHON_TRACE_NOGIL", 1),
                ("CYTHON_USE_SYS_MONITORING", 0),
            ]

    def _get_pth_files(self, ext):
        if ext.name == "mpi4py.MPI" and sys.platform == "win32":
            confdir = os.path.dirname(__file__)
            topdir = os.path.dirname(confdir)
            srcdir = os.path.join(topdir, "src")
            dstdir = self.build_lib
            for pthfile in ("_mpi_dll_path.py", "mpi.pth"):
                source = os.path.join(srcdir, pthfile)
                target = os.path.join(dstdir, pthfile)
                if os.path.exists(source):
                    yield (source, target)

    def build_extension(self, ext):
        fullname = self.get_ext_fullname(ext.name)
        filename = os.path.join(
            self.build_lib, self.get_ext_filename(fullname)
        )
        depends = ext.sources + ext.depends
        if not (self.force or newer_group(depends, filename, "newer")):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        #
        self.config_extension(ext)
        cmd_build_ext.build_ext.build_extension(self, ext)
        #
        for source, target in self._get_pth_files(ext):
            log.info("writing %s", target)
            copy_file(source, target, verbose=False, dry_run=self.dry_run)

    def get_outputs(self):
        outputs = cmd_build_ext.build_ext.get_outputs(self)
        for ext in self.extensions:
            for _, target in self._get_pth_files(ext):
                outputs.append(target)
        return outputs


# Command class to build executables


class build_exe(build_ext):
    description = "build binary executable components"

    user_options = [
        ("build-exe=", None, "build directory for executable components"),
        *build_ext.user_options,
    ]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.build_base = None
        self.build_exe = None
        self.inplace = None

    def finalize_options(self):
        build_ext.finalize_options(self)
        self.configure = None
        self.set_undefined_options(
            "build", ("build_base", "build_base"), ("build_lib", "build_exe")
        )
        self.set_undefined_options("build_ext", ("inplace", "inplace"))
        self.executables = self.distribution.executables
        # NOTE: This is a hack
        self.extensions = self.distribution.executables
        self.get_ext_filename = self.get_exe_filename
        self.check_extensions_list = self.check_executables_list
        self.build_extension = self.build_executable
        self.copy_extensions_to_source = self.copy_executables_to_source
        self.build_lib = self.build_exe

    def get_exe_filename(self, exe_name):
        exe_ext = sysconfig.get_config_var("EXE") or ""
        return exe_name + exe_ext

    def check_executables_list(self, executables):
        ListType, TupleType = type([]), type(())
        if type(executables) is not ListType:
            raise DistutilsSetupError(
                "'executables' option must be a list of Executable instances"
            )
        for exe in executables:
            if not isinstance(exe, Executable):
                raise DistutilsSetupError(
                    "'executables' items must be Executable instances"
                )
            if exe.sources is None or type(exe.sources) not in (
                ListType,
                TupleType,
            ):
                raise DistutilsSetupError(
                    f"in 'executables' option (executable '{exe.name}'), "
                    "'sources' must be present and must be "
                    "a list of source filenames"
                )

    def get_exe_fullpath(self, exe, build_dir=None):
        build_dir = build_dir or self.build_exe
        package_dir = (exe.package or "").split(".")
        dest_dir = convert_path(exe.dest_dir or "")
        output_dir = os.path.join(build_dir, *[*package_dir, dest_dir])
        exe_filename = self.get_exe_filename(exe.name)
        return os.path.join(output_dir, exe_filename)

    def config_executable(self, exe):
        build_ext.config_extension(self, exe)

    def build_executable(self, exe):
        sources = list(exe.sources)
        depends = list(exe.depends)
        exe_fullpath = self.get_exe_fullpath(exe)
        depends = sources + depends
        if not (self.force or newer_group(depends, exe_fullpath, "newer")):
            log.debug("skipping '%s' executable (up-to-date)", exe.name)
            return

        self.config_executable(exe)
        log.info("building '%s' executable", exe.name)

        # Next, compile the source code to object files.

        # Two possible sources for extra compiler arguments:
        #   - 'extra_compile_args' in Extension object
        #   - 'define_macros' or 'undef_macros' in Extension object
        #   - CFLAGS environment variable (not particularly
        #     elegant, but people seem to expect it and I
        #     guess it's useful)
        # The environment variable should take precedence, and
        # any sensible compiler will give precedence to later
        # command line args.  Hence we combine them in order:
        extra_args = exe.extra_compile_args[:]

        macros = exe.define_macros[:]
        for undef in exe.undef_macros:
            macros.append((undef,))

        objects = self.compiler.compile(
            sources,
            output_dir=self.build_temp,
            macros=macros,
            include_dirs=exe.include_dirs,
            debug=self.debug,
            extra_postargs=extra_args,
            depends=exe.depends,
        )
        self._built_objects = objects[:]

        # Now link the object files together into a "shared object" --
        # of course, first we have to figure out all the other things
        # that go into the mix.
        if exe.extra_objects:
            objects.extend(exe.extra_objects)
        extra_args = exe.extra_link_args[:]
        # Get special linker flags for building a executable with
        # bundled Python library, also fix location of needed
        # python.exp file on AIX
        ldflags = sysconfig.get_config_var("PY_LDFLAGS") or ""
        linkshared = sysconfig.get_config_var("LINKFORSHARED") or ""
        linkshared = linkshared.replace("-Xlinker ", "-Wl,")
        if sys.platform == "darwin":  # fix wrong framework paths
            fwkprefix = sysconfig.get_config_var("PYTHONFRAMEWORKPREFIX")
            fwkdir = sysconfig.get_config_var("PYTHONFRAMEWORKDIR")
            if fwkprefix and fwkdir and fwkdir != "no-framework":
                for flag in shlex.split(linkshared):
                    if flag.startswith(fwkdir):
                        fwkpath = os.path.join(fwkprefix, flag)
                        linkshared = linkshared.replace(flag, fwkpath)
        if sys.platform.startswith("aix"):
            python_lib = sysconfig.get_python_lib(standard_lib=1)
            python_exp = os.path.join(python_lib, "config", "python.exp")
            linkshared = linkshared.replace("Modules/python.exp", python_exp)
        # Detect target language, if not provided
        language = exe.language or self.compiler.detect_language(sources)
        self.compiler.link(
            self.compiler.EXECUTABLE,
            objects,
            exe_fullpath,
            output_dir=None,
            libraries=self.get_libraries(exe),
            library_dirs=exe.library_dirs,
            runtime_library_dirs=exe.runtime_library_dirs,
            extra_preargs=shlex.split(ldflags) + shlex.split(linkshared),
            extra_postargs=extra_args,
            debug=self.debug,
            target_lang=language,
        )

    def copy_executables_to_source(self):
        build_py = self.get_finalized_command("build_py")
        root_dir = build_py.get_package_dir("")
        for exe in self.executables:
            src = self.get_exe_fullpath(exe)
            dest = self.get_exe_fullpath(exe, root_dir)
            self.mkpath(os.path.dirname(dest))
            copy_file(src, dest, verbose=self.verbose, dry_run=self.dry_run)

    def get_outputs(self):
        outputs = []
        for exe in self.executables:
            outputs.append(self.get_exe_fullpath(exe))
        return outputs


class install(cmd_install.install):
    def initialize_options(self):
        with warnings.catch_warnings():
            if setuptools:
                category = setuptools.SetuptoolsDeprecationWarning
                warnings.simplefilter("ignore", category)
            cmd_install.install.initialize_options(self)
        self.old_and_unmanageable = True

    def run(self):
        cmd_install.install.run(self)

    def has_lib(self):
        return cmd_install.install.has_lib(self) and self.has_exe()

    def has_exe(self):
        return self.distribution.has_executables()

    sub_commands = [
        *cmd_install.install.sub_commands,
        ("install_exe", has_exe),
    ]

    # NOTE: disable install_exe subcommand !!!
    del sub_commands[-1]


class install_lib(cmd_install_lib.install_lib):
    def get_outputs(self):
        outputs = cmd_install_lib.install_lib.get_outputs(self)
        for build_cmd, build_dir in (("build_exe", "build_exe"),):
            cmd_obj = self.get_finalized_command(build_cmd)
            build_files = cmd_obj.get_outputs()
            exe_outputs = self._mutate_outputs(
                self.distribution.has_executables(),
                build_cmd,
                build_dir,
                self.install_dir,
            )
            for src, dest in zip(build_files, exe_outputs):
                if os.path.exists(src):
                    outputs.append(dest)
        return outputs


class install_data(cmd_install_data.install_data):
    def finalize_options(self):
        self.set_undefined_options(
            "install",
            ("install_lib", "install_dir"),
            ("root", "root"),
            ("force", "force"),
        )


class install_exe(cmd_install_lib.install_lib):
    description = "install binary executable components"

    user_options = [
        ("install-dir=", "d", "directory to install to"),
        ("build-dir=", "b", "build directory (where to install from)"),
        ("force", "f", "force installation (overwrite existing files)"),
        ("skip-build", None, "skip the build steps"),
    ]

    boolean_options = ["force", "skip-build"]
    negative_opt = {}

    def initialize_options(self):
        self.install_dir = None
        self.build_dir = None
        self.force = 0
        self.skip_build = None

    def finalize_options(self):
        self.set_undefined_options("build_exe", ("build_exe", "build_dir"))
        self.set_undefined_options(
            "install",
            ("force", "force"),
            ("skip_build", "skip_build"),
            ("install_scripts", "install_dir"),
        )

    def run(self):
        self.build()
        self.install()

    def build(self):
        if not self.skip_build:
            if self.distribution.has_executables():
                self.run_command("build_exe")

    def install(self):
        self.outfiles = []
        if self.distribution.has_executables():
            build_exe = self.get_finalized_command("build_exe")
            for exe in build_exe.executables:
                exe_fullpath = build_exe.get_exe_fullpath(exe)
                exe_filename = os.path.basename(exe_fullpath)
                if exe_filename.startswith("python-") and os.name == "posix":
                    x, y = sys.version_info[:2]
                    install_name = exe_filename.replace(
                        "python-", f"python{x}.{y}-"
                    )
                    link = None
                else:
                    install_name = exe_filename
                    link = None
                source = exe_fullpath
                target = os.path.join(self.install_dir, install_name)
                self.mkpath(self.install_dir)
                out, _done = self.copy_file(source, target, link=link)
                self.outfiles.append(out)

    def get_outputs(self):
        return self.outfiles

    def get_inputs(self):
        inputs = []
        if self.distribution.has_executables():
            build_exe = self.get_finalized_command("build_exe")
            inputs.extend(build_exe.get_outputs())
        return inputs


class clean(cmd_clean.clean):
    description = "clean up temporary files from 'build' command"
    user_options = [
        *cmd_clean.clean.user_options[:2],
        (
            "build-exe=",
            None,
            (
                "build directory for executable components "
                "[default: 'build_exe.build-exe']"
            ),
        ),
        *cmd_clean.clean.user_options[2:],
    ]

    def initialize_options(self):
        cmd_clean.clean.initialize_options(self)
        self.build_exe = None

    def finalize_options(self):
        cmd_clean.clean.finalize_options(self)
        self.set_undefined_options("build_exe", ("build_exe", "build_exe"))

    def run(self):
        from distutils.dir_util import remove_tree

        # remove the build/temp.<plat> directory
        # (unless it's already gone)
        if os.path.exists(self.build_temp):
            remove_tree(self.build_temp, dry_run=self.dry_run)
        else:
            log.debug("'%s' does not exist -- can't clean it", self.build_temp)

        if self.all:
            # remove build directories
            for directory in (
                self.build_lib,
                self.build_exe,
                self.build_scripts,
                self.bdist_base,
            ):
                if os.path.exists(directory):
                    remove_tree(directory, dry_run=self.dry_run)
                else:
                    log.debug(
                        "'%s' does not exist -- can't clean it", directory
                    )

        # just for the heck of it, try to remove the base build directory:
        # we might have emptied it right now, but if not we don't care
        if not self.dry_run:
            try:
                os.rmdir(self.build_base)
                log.info("removing '%s'", self.build_base)
            except OSError:
                pass

        if self.all:
            # remove the <package>.egg_info directory
            try:
                egg_info = self.get_finalized_command("egg_info").egg_info
                if os.path.exists(egg_info):
                    remove_tree(egg_info, dry_run=self.dry_run)
                else:
                    log.debug(
                        "'%s' does not exist -- can't clean it", egg_info
                    )
            except DistutilsError:
                pass


# -----------------------------------------------------------------------------

if setuptools:
    with contextlib.suppress(Exception):
        from setuptools.command import egg_info as mod_egg_info

        class FileList(mod_egg_info.FileList):
            def process_template_line(self, line):
                level = log.set_threshold(log.ERROR)
                try:
                    super().process_template_line(line)
                finally:
                    log.set_threshold(level)

        mod_egg_info.FileList = FileList

# -----------------------------------------------------------------------------

# Support for Reproducible Builds
# https://reproducible-builds.org/docs/source-date-epoch/

timestamp = os.environ.get("SOURCE_DATE_EPOCH")
if timestamp is not None:
    import distutils.archive_util as archive_util
    import stat
    import tarfile
    import time

    timestamp = float(max(int(timestamp), 0))

    class Time:
        @staticmethod
        def time():
            return timestamp

        @staticmethod
        def localtime(_=None):
            return time.localtime(timestamp)

    class TarInfoMode:
        def __get__(self, obj, objtype=None):
            return obj._mode

        def __set__(self, obj, stmd):
            ifmt = stat.S_IFMT(stmd)
            mode = stat.S_IMODE(stmd) & 0o7755
            obj._mode = ifmt | mode

    class TarInfoAttr:
        def __init__(self, value):
            self.value = value

        def __get__(self, obj, objtype=None):
            return self.value

        def __set__(self, obj, value):
            pass

    class TarInfo(tarfile.TarInfo):
        mode = TarInfoMode()
        mtime = TarInfoAttr(timestamp)
        uid = TarInfoAttr(0)
        gid = TarInfoAttr(0)
        uname = TarInfoAttr("")
        gname = TarInfoAttr("")

    def make_tarball(*args, **kwargs):
        tarinfo_orig = tarfile.TarFile.tarinfo
        try:
            tarfile.time = Time()
            tarfile.TarFile.tarinfo = TarInfo
            return archive_util.make_tarball(*args, **kwargs)
        finally:
            tarfile.time = time
            tarfile.TarFile.tarinfo = tarinfo_orig

    archive_util.ARCHIVE_FORMATS["gztar"] = (
        make_tarball,
        *archive_util.ARCHIVE_FORMATS["gztar"][1:],
    )

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

# CUDA array interface for interoperating Python CUDA GPU libraries
# See https://numba.pydata.org/numba-doc/latest/cuda/cuda_array_interface.html

cdef inline int cuda_is_contig(
    tuple shape,
    tuple strides,
    Py_ssize_t itemsize,
    char order,
) except -1:
    cdef Py_ssize_t i, ndim = len(shape)
    cdef Py_ssize_t start, step, index, dim, size = itemsize
    if order == c'F':
        start = 0
        step = 1
    else:
        start = ndim - 1
        step = -1
    for i in range(ndim):
        index = start + step * i
        dim = <Py_ssize_t>shape[index]
        if dim > 1 and size != <Py_ssize_t>strides[index]:
            return 0
        size *= dim
    return 1

cdef inline char* cuda_get_format(
    char typekind,
    Py_ssize_t itemsize,
) noexcept nogil:
    if typekind == c'b':
        if itemsize == sizeof(char): return b"?"
    if typekind == c'i':
        if itemsize == sizeof(char):      return b"b"
        if itemsize == sizeof(short):     return b"h"
        if itemsize == sizeof(int):       return b"i"
        if itemsize == sizeof(long):      return b"l"
        if itemsize == sizeof(long long): return b"q"  # ~> long
    if typekind == c'u':
        if itemsize == sizeof(char):      return b"B"
        if itemsize == sizeof(short):     return b"H"
        if itemsize == sizeof(int):       return b"I"
        if itemsize == sizeof(long):      return b"L"
        if itemsize == sizeof(long long): return b"Q"  # ~> long
    if typekind == c'f':
        if itemsize == sizeof(float)//2:    return b"e"
        if itemsize == sizeof(float):       return b"f"
        if itemsize == sizeof(double):      return b"d"
        if itemsize == sizeof(long double): return b"g"
    if typekind == c'c':
        if itemsize == 2*sizeof(float)//2:    return b"Ze"
        if itemsize == 2*sizeof(float):       return b"Zf"
        if itemsize == 2*sizeof(double):      return b"Zd"
        if itemsize == 2*sizeof(long double): return b"Zg"
    return BYTE_FMT

# -----------------------------------------------------------------------------

cdef int Py_CheckCAIBuffer(object obj) noexcept:
    try:    return <bint>hasattr(obj, '__cuda_array_interface__')
    except: return 0  # ~> uncovered  # noqa

cdef int Py_GetCAIBuffer(object obj, Py_buffer *view, int flags) except -1:
    cdef dict cuda_array_interface
    cdef tuple data
    cdef str   typestr
    cdef tuple shape
    cdef tuple strides
    cdef list descr
    cdef object dev_ptr, mask
    cdef void *buf = NULL
    cdef bint readonly = 0
    cdef Py_ssize_t s, size = 1
    cdef Py_ssize_t itemsize = 1
    cdef char *format = BYTE_FMT
    cdef char byteorder = c'|'
    cdef char typekind = c'u'

    try:
        cuda_array_interface = obj.__cuda_array_interface__
    except AttributeError:
        raise NotImplementedError("missing CUDA array interface")

    # mandatory
    data = cuda_array_interface['data']
    typestr = cuda_array_interface['typestr']
    shape = cuda_array_interface['shape']

    # optional
    strides = cuda_array_interface.get('strides')
    descr = cuda_array_interface.get('descr')
    mask = cuda_array_interface.get('mask')

    dev_ptr, readonly = data
    for s in shape: size *= s
    if dev_ptr is None and size == 0: dev_ptr = 0
    buf = PyLong_AsVoidPtr(dev_ptr)
    byteorder = <char>ord(typestr[0:1])
    typekind = <char>ord(typestr[1:2])
    itemsize = <Py_ssize_t>int(typestr[2:])
    format = cuda_get_format(typekind, itemsize)

    if (flags & PyBUF_FORMAT) == PyBUF_FORMAT:
        if byteorder == c'<':  # little-endian
            if not is_little_endian():
                raise BufferError(                  # ~> big-endian
                    f"__cuda_array_interface__: "   # ~> big-endian
                    f"typestr {typestr!r} "         # ~> big-endian
                    f"with non-native byte order")  # ~> big-endian
        elif byteorder == c'>':  # big-endian
            if not is_big_endian():
                raise BufferError(                  # ~> little-endian
                    f"__cuda_array_interface__: "   # ~> little-endian
                    f"typestr {typestr!r} "         # ~> little-endian
                    f"with non-native byte order")  # ~> little-endian
        elif byteorder != c'|':
            raise BufferError(
                f"__cuda_array_interface__: "
                f"typestr {typestr!r} "
                f"with unrecognized byte order")
    if mask is not None:
        raise BufferError(
            "__cuda_array_interface__: "
            "cannot handle masked arrays"
        )
    if size < 0:
        raise BufferError(
            f"__cuda_array_interface__: "
            f"buffer with negative size "
            f"(shape:{shape}, size:{size})"
        )
    if (
        strides is not None and
        not cuda_is_contig(shape, strides, itemsize, c'C') and
        not cuda_is_contig(shape, strides, itemsize, c'F')
    ):
        raise BufferError(
            f"__cuda_array_interface__: "
            f"buffer is not contiguous "
            f"(shape:{shape}, strides:{strides}, itemsize:{itemsize})"
        )
    if readonly and ((flags & PyBUF_WRITABLE) == PyBUF_WRITABLE):
        raise BufferError(
            f"__cuda_array_interface__: "
            f"buffer is not writable"
        )
    if  descr is not None and (len(descr) != 1 or descr[0] != ('', typestr)):
        PyErr_WarnFormat(
            RuntimeWarning, 1,
            b"__cuda_array_interface__: %s",
            b"ignoring 'descr' key",
        )

    PyBuffer_FillInfo(view, obj, buf, size*itemsize, readonly, flags)

    if (flags & PyBUF_FORMAT) == PyBUF_FORMAT:
        view.format = format
        if view.format != BYTE_FMT:
            view.itemsize = itemsize

    return <int> kDLCUDA

# -----------------------------------------------------------------------------



import ctypes
import operator
import os
import sys
import weakref

import mpiunittest as unittest

from mpi4py import MPI


class TestObjModel(unittest.TestCase):
    #
    objects = [
        MPI.Status(),
        MPI.DATATYPE_NULL,
        MPI.REQUEST_NULL,
        MPI.INFO_NULL,
        MPI.ERRHANDLER_NULL,
        MPI.SESSION_NULL,
        MPI.GROUP_NULL,
        MPI.WIN_NULL,
        MPI.OP_NULL,
        MPI.FILE_NULL,
        MPI.MESSAGE_NULL,
        MPI.COMM_NULL,
    ]

    def testEq(self):
        for i, obj1 in enumerate(self.objects):
            objects = self.objects[:]
            obj2 = objects[i]
            self.assertTrue(bool(obj1 == obj2))
            self.assertFalse(bool(obj1 != obj2))
            del objects[i]
            for obj2 in objects:
                self.assertTrue(bool(obj1 != obj2))
                self.assertTrue(bool(obj2 != obj1))
                self.assertFalse(bool(obj1 == obj2))
                self.assertFalse(bool(obj2 == obj1))
            self.assertFalse(bool(None is obj1))
            self.assertFalse(bool(obj1 is None))
            self.assertFalse(bool(obj1 is True))
            self.assertFalse(bool(obj1 is False))
            self.assertFalse(bool(obj1 == 12345))
            self.assertFalse(bool(obj1 == "abc"))
            self.assertFalse(bool(obj1 == [123]))
            self.assertFalse(bool(obj1 == (1, 2)))
            self.assertFalse(bool(obj1 == {0: 0}))
            self.assertFalse(bool(obj1 == set()))

    def testNe(self):
        for i, obj1 in enumerate(self.objects):
            objects = self.objects[:]
            obj2 = objects[i]
            self.assertFalse(bool(obj1 != obj2))
            del objects[i]
            for obj2 in objects:
                self.assertTrue(bool(obj1 != obj2))
            self.assertTrue(bool(None is not obj1))
            self.assertTrue(bool(obj1 is not None))
            self.assertTrue(bool(obj1 is not True))
            self.assertTrue(bool(obj1 is not False))
            self.assertTrue(bool(obj1 != 12345))
            self.assertTrue(bool(obj1 != "abc"))
            self.assertTrue(bool(obj1 != [123]))
            self.assertTrue(bool(obj1 != (1, 2)))
            self.assertTrue(bool(obj1 != {0: 0}))
            self.assertTrue(bool(obj1 != set()))

    def testCmp(self):
        for obj in self.objects:
            for binop in ("lt", "le", "gt", "ge"):
                binop = getattr(operator, binop)
                with self.assertRaises(TypeError):
                    binop(obj, obj)

    def testBool(self):
        for obj in self.objects[1:]:
            self.assertFalse(not not obj)
            self.assertTrue(not obj)
            self.assertFalse(obj)

    def testReduce(self):
        import copy
        import pickle

        def functions(_obj):
            for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
                yield lambda ob, _=protocol: pickle.loads(pickle.dumps(ob, _))
            yield copy.copy
            yield copy.deepcopy

        for obj in self.objects:
            for copier in functions(obj):
                dup = copier(obj)
                self.assertIs(type(dup), type(obj))
                if isinstance(obj, MPI.Status):
                    self.assertIsNot(dup, obj)
                else:
                    self.assertIs(dup, obj)
                cls = type(obj)
                dup = copier(cls(obj))
                self.assertIs(type(dup), cls)
                self.assertIsNot(dup, obj)
                cls = type(f"My{type(obj).__name__}", (type(obj),), {})
                main = __import__("__main__")
                cls.__module__ = main.__name__
                setattr(main, cls.__name__, cls)
                dup = copier(cls(obj))
                delattr(main, cls.__name__)
                self.assertIs(type(dup), cls)
                self.assertIsNot(dup, obj)

    def testHash(self):
        for obj in self.objects:
            with self.assertRaises(TypeError):
                hash(obj)

    def testInit(self):
        for i, obj in enumerate(self.objects):
            klass = type(obj)
            new = klass()
            self.assertEqual(new, obj)
            new = klass(obj)
            self.assertEqual(new, obj)

            objects = self.objects[:]
            del objects[i]
            for other in objects:
                with self.assertRaises(TypeError):
                    klass(other)

            with self.assertRaises(TypeError):
                klass(1234)
            with self.assertRaises(TypeError):
                klass("abc")

    def testWeakRef(self):
        for obj in self.objects:
            wr = weakref.ref(obj)
            self.assertIs(wr(), obj)
            self.assertIn(wr, weakref.getweakrefs(obj))
            wr = weakref.proxy(obj)
            self.assertIn(wr, weakref.getweakrefs(obj))

    def testHandle(self):
        objects = self.objects[:]
        objects += [
            MPI.INT,
            MPI.FLOAT,
            MPI.Request(MPI.REQUEST_NULL),
            MPI.Prequest(MPI.REQUEST_NULL),
            MPI.Grequest(MPI.REQUEST_NULL),
            MPI.INFO_ENV,
            MPI.GROUP_EMPTY,
            MPI.ERRORS_RETURN,
            MPI.ERRORS_ABORT,
            MPI.ERRORS_ARE_FATAL,
            MPI.COMM_SELF,
            MPI.COMM_WORLD,
        ]
        for obj in objects:
            if isinstance(obj, MPI.Status):
                continue
            self.assertGreaterEqual(obj.handle, 0)
            newobj = type(obj).fromhandle(obj.handle)
            self.assertEqual(newobj, obj)
            self.assertEqual(type(newobj), type(obj))
            self.assertEqual(newobj.handle, obj.handle)
            with self.assertRaises(AttributeError):
                newobj.handle = None
            with self.assertRaises(AttributeError):
                newobj.handle = obj.handle
            with self.assertRaises(AttributeError):
                del newobj.handle

    def testSafeFreeNull(self):
        objects = self.objects[:]
        for obj in objects:
            if isinstance(obj, MPI.Status):
                continue
            obj.free()
            self.assertFalse(obj)
            obj.free()
            self.assertFalse(obj)

    def testSafeFreeConstant(self):
        objects = [
            MPI.INT,
            MPI.LONG,
            MPI.FLOAT,
            MPI.DOUBLE,
            MPI.INFO_ENV,
            MPI.SUM,
            MPI.PROD,
            MPI.GROUP_EMPTY,
            MPI.ERRORS_ABORT,
            MPI.ERRORS_ARE_FATAL,
            MPI.ERRORS_RETURN,
            MPI.MESSAGE_NO_PROC,
            MPI.COMM_SELF,
            MPI.COMM_WORLD,
        ]
        for obj in filter(None, objects):
            self.assertTrue(obj)
            for _ in range(3):
                obj.free()
                self.assertTrue(obj)
            if not isinstance(obj, MPI.Errhandler):
                clon = type(obj)(obj)
                self.assertTrue(clon)
                for _ in range(3):
                    clon.free()
                    self.assertFalse(clon)
            if hasattr(obj, "Dup"):
                self.assertTrue(obj)
                dup = obj.Dup()
                self.assertTrue(dup)
                for _ in range(3):
                    dup.free()
                    self.assertFalse(dup)
            self.assertTrue(obj)
            for _ in range(3):
                obj.free()
                self.assertTrue(obj)

    def testSafeFreeCreated(self):
        objects = [
            MPI.COMM_SELF.Isend((None, 0, MPI.BYTE), MPI.PROC_NULL),
            MPI.Op.Create(lambda *_: None),
            MPI.COMM_SELF.Get_group(),
            MPI.COMM_SELF.Get_errhandler(),
        ]
        try:
            objects += [MPI.Info.Create()]
        except (NotImplementedError, MPI.Exception):
            pass
        if os.name == "posix":
            try:
                objects += [MPI.File.Open(MPI.COMM_SELF, "/dev/null")]
            except NotImplementedError:
                pass
        try:
            objects += [MPI.Win.Create(MPI.BOTTOM)]
        except (NotImplementedError, MPI.Exception):
            pass
        try:
            objects += [MPI.Session.Init()]
        except NotImplementedError:
            pass
        for obj in objects:
            self.assertTrue(obj)
            for _ in range(3):
                obj.free()
                self.assertFalse(obj)

    def testConstants(self):
        import pickle

        names = (
            "BOTTOM",
            "IN_PLACE",
            "BUFFER_AUTOMATIC",
        )
        for name in names:
            constant = getattr(MPI, name)
            self.assertTrue(operator.eq(constant, constant))
            self.assertTrue(operator.eq(constant, int(constant)))
            self.assertTrue(operator.ne(constant, None))
            self.assertFalse(operator.ne(constant, constant))
            self.assertFalse(operator.ne(constant, int(constant)))
            self.assertFalse(operator.eq(constant, None))
            self.assertEqual(constant, int(constant))
            self.assertEqual(hash(constant), hash(int(constant)))
            self.assertEqual(bool(constant), bool(int(constant)))
            self.assertEqual(type(constant)(), constant)
            self.assertEqual(memoryview(constant).nbytes, 0)
            self.assertEqual(MPI.Get_address(constant), constant)
            if sys.implementation.name != "pypy":
                self.assertIsNone(memoryview(constant).obj)
            with self.assertRaises(TypeError):
                constant + 1
            self.assertEqual(repr(constant), name)
            self.assertEqual(constant.__reduce__(), name)
            for protocol in range(pickle.HIGHEST_PROTOCOL):
                value = pickle.loads(pickle.dumps(constant, protocol))
                self.assertIs(type(value), type(constant))
                self.assertEqual(value, constant)
        constant = MPI.BOTTOM
        self.assertEqual(operator.index(constant), int(constant))

    def testSizeOf(self):
        for obj in self.objects:
            n1 = MPI._sizeof(obj)
            n2 = MPI._sizeof(type(obj))
            self.assertEqual(n1, n2)
        with self.assertRaises(TypeError):
            MPI._sizeof(None)

    def testAddressOf(self):
        for obj in self.objects:
            addr = MPI._addressof(obj)
            self.assertNotEqual(addr, 0)
        with self.assertRaises(TypeError):
            MPI._addressof(None)

    def testAHandleOf(self):
        for obj in self.objects:
            hdl = MPI._handleof(obj)
            self.assertGreaterEqual(hdl, 0)
        with self.assertRaises(TypeError):
            MPI._handleof(None)

    @unittest.skipUnless(sys.implementation.name == "cpython", "cpython")
    @unittest.skipUnless(hasattr(MPI, "__pyx_capi__"), "cython")
    def testCAPI(self):
        status = MPI.Status()
        status.source = 0
        status.tag = 1
        status.error = MPI.ERR_OTHER
        extra_objects = [
            status,
            MPI.INT,
            MPI.SUM,
            MPI.INFO_ENV,
            MPI.MESSAGE_NO_PROC,
            MPI.ERRORS_RETURN,
            MPI.GROUP_EMPTY,
            MPI.COMM_SELF,
        ]

        pyapi = ctypes.pythonapi
        PyCapsule_GetPointer = pyapi.PyCapsule_GetPointer
        PyCapsule_GetPointer.restype = ctypes.c_void_p
        PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]

        pyx_capi = MPI.__pyx_capi__

        for obj in self.objects + extra_objects:
            cls = type(obj)
            if issubclass(cls, MPI.Comm):
                cls = MPI.Comm
            typename = cls.__name__
            modifier = ""
            if isinstance(obj, MPI.Status):
                mpi_type = ctypes.c_void_p
                modifier = " *"
            elif MPI._sizeof(cls) == ctypes.sizeof(ctypes.c_uint32):
                mpi_type = ctypes.c_uint32
            elif MPI._sizeof(cls) == ctypes.sizeof(ctypes.c_uint64):
                mpi_type = ctypes.c_uint64

            new_functype = ctypes.PYFUNCTYPE(ctypes.py_object, mpi_type)
            get_functype = ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.py_object)
            new_capsule = pyx_capi[f"PyMPI{typename}_New"]
            get_capsule = pyx_capi[f"PyMPI{typename}_Get"]
            new_signature = f"PyObject *(MPI_{typename}{modifier})".encode()
            get_signature = f"MPI_{typename} *(PyObject *)".encode()
            PyCapsule_GetPointer.restype = new_functype
            pympi_new = PyCapsule_GetPointer(new_capsule, new_signature)
            PyCapsule_GetPointer.restype = get_functype
            pympi_get = PyCapsule_GetPointer(get_capsule, get_signature)
            PyCapsule_GetPointer.restype = ctypes.c_void_p

            objptr = pympi_get(obj)
            if isinstance(obj, MPI.Status):
                newarg = objptr
            else:
                newarg = mpi_type.from_address(objptr).value
            self.assertEqual(objptr, MPI._addressof(obj))
            self.assertEqual(newarg, MPI._handleof(obj))
            newobj = pympi_new(newarg)
            self.assertIs(type(newobj), type(obj))
            self.assertEqual(newobj, obj)


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI

try:
    import array
except ImportError:
    array = None


class TestBuffer(unittest.TestCase):
    #
    def testNewEmpty(self):
        buffer = MPI.buffer
        buf = buffer()
        self.assertEqual(buf.address, 0)
        self.assertIsNone(buf.obj)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)
        self.assertEqual(buf.format, "B")
        self.assertEqual(buf.itemsize, 1)
        self.assertEqual(len(buf), 0)
        buf[:] = 0
        buf[:] = buffer()
        m = memoryview(buf)
        self.assertEqual(m.format, "B")
        self.assertEqual(m.itemsize, 1)
        self.assertEqual(m.ndim, 1)
        self.assertIs(m.readonly, False)
        self.assertEqual(m.shape, (0,))
        self.assertEqual(m.strides, (1,))
        self.assertEqual(m.tobytes(), b"")
        self.assertEqual(m.tolist(), [])
        buf.release()
        self.assertEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)

    def testNewBad(self):
        buffer = MPI.buffer
        for obj in (None, 0, 0.0, [], (), []):
            self.assertRaises(TypeError, buffer, obj)

    def testNewBytes(self):
        buffer = MPI.buffer
        obj = b"abc"
        buf = buffer(obj)
        self.assertEqual(buf.obj, obj)
        self.assertEqual(buf.nbytes, len(obj))
        self.assertIs(buf.readonly, True)
        with self.assertRaises(TypeError):
            buf[:] = 0

    def testNewBytearray(self):
        buffer = MPI.buffer
        obj = bytearray([1, 2, 3])
        buf = buffer(obj)
        self.assertEqual(buf.obj, obj)
        self.assertEqual(buf.nbytes, len(obj))
        self.assertFalse(buf.readonly)
        with self.assertRaises(ValueError):
            buf[0:1] = buf[1:3]

    @unittest.skipIf(array is None, "array")
    def testNewArray(self):
        buffer = MPI.buffer
        obj = array.array("i", [1, 2, 3])
        buf = buffer(obj)
        self.assertEqual(buf.obj, obj)
        self.assertEqual(buf.nbytes, len(obj) * obj.itemsize)
        self.assertFalse(buf.readonly)

    def testAllocate(self):
        buffer = MPI.buffer
        for size in (0, 1, 2):
            buf = buffer.allocate(size)
            self.assertEqual(buf.nbytes, size)
            self.assertNotEqual(buf.address, 0)
            view = memoryview(buf.obj)
            self.assertEqual(buf.nbytes, view.nbytes)
        for clear in (False, True):
            buf = buffer.allocate(1024, clear)
            self.assertEqual(buf.nbytes, 1024)
            self.assertNotEqual(buf.address, 0)
            if clear:
                self.assertEqual(buf[0], 0)
                self.assertEqual(buf[-1], 0)
        self.assertRaises(TypeError, buffer.allocate, None)
        self.assertRaises(ValueError, buffer.allocate, -1)

    def testFromBufferBad(self):
        buffer = MPI.buffer
        for obj in (None, 0, 0.0, [], (), []):
            self.assertRaises(TypeError, buffer.frombuffer, obj)

    def testFromBufferBytes(self):
        buffer = MPI.buffer
        buf = buffer.frombuffer(b"abc", readonly=True)
        self.assertNotEqual(buf.address, 0)
        self.assertEqual(type(buf.obj), bytes)
        self.assertEqual(buf.obj, b"abc")
        self.assertEqual(buf.nbytes, 3)
        self.assertTrue(buf.readonly)
        self.assertEqual(buf.format, "B")
        self.assertEqual(buf.itemsize, 1)
        self.assertEqual(len(buf), 3)
        m = memoryview(buf)
        self.assertEqual(m.format, "B")
        self.assertEqual(m.itemsize, 1)
        self.assertEqual(m.ndim, 1)
        self.assertTrue(m.readonly)
        self.assertEqual(m.shape, (3,))
        self.assertEqual(m.strides, (1,))
        self.assertEqual(m.tobytes(), b"abc")
        self.assertEqual(m.tolist(), [ord(c) for c in "abc"])
        buf.release()
        self.assertEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)

    @unittest.skipIf(array is None, "array")
    def testFromBufferArrayRO(self):
        buffer = MPI.buffer
        obj = array.array("B", [1, 2, 3])
        buf = buffer.frombuffer(obj, readonly=True)
        self.assertNotEqual(buf.address, 0)
        self.assertEqual(type(buf.obj), array.array)
        self.assertEqual(buf.nbytes, 3)
        self.assertTrue(buf.readonly)
        self.assertEqual(buf.format, "B")
        self.assertEqual(buf.itemsize, 1)
        self.assertEqual(len(buf), 3)
        m = memoryview(buf)
        self.assertEqual(m.format, "B")
        self.assertEqual(m.itemsize, 1)
        self.assertEqual(m.ndim, 1)
        self.assertTrue(m.readonly)
        self.assertEqual(m.shape, (3,))
        self.assertEqual(m.strides, (1,))
        self.assertEqual(m.tobytes(), b"\1\2\3")
        self.assertEqual(m.tolist(), [1, 2, 3])
        buf.release()
        self.assertEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)

    @unittest.skipIf(array is None, "array")
    def testFromBufferArrayRW(self):
        buffer = MPI.buffer
        obj = array.array("B", [1, 2, 3])
        buf = buffer.frombuffer(obj, readonly=False)
        self.assertNotEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 3)
        self.assertFalse(buf.readonly)
        self.assertEqual(len(buf), 3)
        m = memoryview(buf)
        self.assertEqual(m.format, "B")
        self.assertEqual(m.itemsize, 1)
        self.assertEqual(m.ndim, 1)
        self.assertFalse(m.readonly)
        self.assertEqual(m.shape, (3,))
        self.assertEqual(m.strides, (1,))
        self.assertEqual(m.tobytes(), b"\1\2\3")
        self.assertEqual(m.tolist(), [1, 2, 3])
        buf[:] = 1
        self.assertEqual(obj, array.array("B", [1] * 3))
        buf[1:] = array.array("B", [7] * 2)
        self.assertEqual(obj, array.array("B", [1, 7, 7]))
        buf[1:2] = array.array("B", [8] * 1)
        self.assertEqual(obj, array.array("B", [1, 8, 7]))
        buf.release()
        self.assertEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)

    @unittest.skipIf(array is None, "array")
    def testFromAddress(self):
        buffer = MPI.buffer
        obj = array.array("B", [1, 2, 3])
        addr, size = obj.buffer_info()
        nbytes = size * obj.itemsize
        buf = buffer.fromaddress(addr, nbytes, readonly=False)
        self.assertNotEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 3)
        self.assertFalse(buf.readonly)
        self.assertEqual(len(buf), 3)
        m = memoryview(buf)
        self.assertEqual(m.format, "B")
        self.assertEqual(m.itemsize, 1)
        self.assertEqual(m.ndim, 1)
        self.assertFalse(m.readonly)
        self.assertEqual(m.shape, (3,))
        self.assertEqual(m.strides, (1,))
        self.assertEqual(m.tobytes(), b"\1\2\3")
        self.assertEqual(m.tolist(), [1, 2, 3])
        buf[:] = 1
        self.assertEqual(obj, array.array("B", [1] * 3))
        buf[1:] = array.array("B", [7] * 2)
        self.assertEqual(obj, array.array("B", [1, 7, 7]))
        buf[1:2] = array.array("B", [8] * 1)
        self.assertEqual(obj, array.array("B", [1, 8, 7]))
        buf.release()
        self.assertEqual(buf.address, 0)
        self.assertEqual(buf.nbytes, 0)
        self.assertFalse(buf.readonly)
        with self.assertRaises(ValueError):
            buffer.fromaddress(addr, -1)
        with self.assertRaises(ValueError):
            buffer.fromaddress(0, 1)

    def testToReadonly(self):
        buffer = MPI.buffer
        obj = bytearray(b"abc")
        buf1 = buffer.frombuffer(obj)
        buf2 = buf1.toreadonly()
        self.assertFalse(buf1.readonly)
        self.assertTrue(buf2.readonly)
        self.assertEqual(buf1.address, buf2.address)
        self.assertEqual(buf1.obj, buf2.obj)
        self.assertEqual(type(buf1.obj), type(buf2.obj))
        self.assertEqual(buf1.nbytes, buf2.nbytes)

    def testCast(self):
        buffer = MPI.buffer
        buf = buffer.allocate(2 * 3 * 4)
        mem = buf.cast("i")
        for i in range(2 * 3):
            mem[i] = i
        mem = buf.cast("i", (2, 3))
        for i in range(2):
            for j in range(3):
                self.assertEqual(mem[i, j], 3 * i + j)
        mem = buf.cast("i", (3, 2))
        for i in range(3):
            for j in range(2):
                self.assertEqual(mem[i, j], 2 * i + j)

    def testSequence(self):
        n = 16
        try:
            mem = MPI.Alloc_mem(n, MPI.INFO_NULL)
        except NotImplementedError:
            self.skipTest("mpi-alloc_mem")
        try:
            self.assertIs(type(mem), MPI.buffer)
            self.assertNotEqual(mem.address, 0)
            self.assertEqual(mem.nbytes, n)
            self.assertFalse(mem.readonly)
            self.assertEqual(len(mem), n)

            def delitem():
                del mem[n]

            def getitem1():
                return mem[n]

            def getitem2():
                return mem[::2]

            def getitem3():
                return mem[None]

            def setitem1():
                mem[n] = 0

            def setitem2():
                mem[::2] = 0

            def setitem3():
                mem[None] = 0

            self.assertRaises(NotImplementedError, delitem)
            self.assertRaises(IndexError, getitem1)
            self.assertRaises(IndexError, getitem2)
            self.assertRaises(TypeError, getitem3)
            self.assertRaises(IndexError, setitem1)
            self.assertRaises(IndexError, setitem2)
            self.assertRaises(TypeError, setitem3)
            for i in range(n):
                mem[i] = i
            for i in range(n):
                self.assertEqual(mem[i], i)
            mem[:] = 0
            for i in range(-n, 0):
                mem[i] = abs(i)
            for i in range(-n, 0):
                self.assertEqual(mem[i], abs(i))
            mem[:] = 0
            for i in range(n):
                self.assertEqual(mem[i], 0)
            mem[:] = 255
            for i in range(n):
                self.assertEqual(mem[i], 255)
            mem[: n // 2] = 1
            mem[n // 2 :] = 0
            for i in range(n // 2):
                self.assertEqual(mem[i], 1)
            for i in range(n // 2, n):
                self.assertEqual(mem[i], 0)
            mem[:] = 0
            mem[1:5] = b"abcd"
            mem[10:13] = b"xyz"
            self.assertEqual(mem[0], 0)
            for i, c in enumerate("abcd"):
                self.assertEqual(mem[1 + i], ord(c))
            for i in range(5, 10):
                self.assertEqual(mem[i], 0)
            for i, c in enumerate("xyz"):
                self.assertEqual(mem[10 + i], ord(c))
            for i in range(13, n):
                self.assertEqual(mem[i], 0)
            self.assertEqual(mem[1:5].tobytes(), b"abcd")
            self.assertEqual(mem[10:13].tobytes(), b"xyz")
        finally:
            MPI.Free_mem(mem)
            self.assertEqual(mem.address, 0)
            self.assertEqual(mem.nbytes, 0)
            self.assertFalse(mem.readonly)

    def testBuffering(self):
        buf = MPI.Alloc_mem((1 << 16) + MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        try:
            with self.catchNotImplementedError(4, 1):
                MPI.Flush_buffer()
            with self.catchNotImplementedError(4, 1):
                MPI.Iflush_buffer().Wait()
        finally:
            oldbuf = MPI.Detach_buffer()
            self.assertEqual(oldbuf.address, buf.address)
            self.assertEqual(oldbuf.nbytes, buf.nbytes)
            MPI.Free_mem(buf)
        if MPI.Get_version() >= (4, 1):
            MPI.Attach_buffer(MPI.BUFFER_AUTOMATIC)
            bufauto = MPI.Detach_buffer()
            self.assertEqual(bufauto, MPI.BUFFER_AUTOMATIC)

    def testAttachBufferReadonly(self):
        buf = MPI.buffer(b"abc")
        self.assertRaises(BufferError, MPI.Attach_buffer, buf)


if not hasattr(MPI, "buffer"):
    unittest.disable(TestBuffer, "mpi4py-buffer")


if __name__ == "__main__":
    unittest.main()



# mpiexec -n 2 python ex-2.34.py

# Use of ready-mode and synchronous-mode

# --------------------------------------------------------------------

from mpi4py import MPI

try:
    import numpy
except ImportError:
    raise SystemExit from None

if MPI.COMM_WORLD.Get_size() < 2:
    raise SystemExit from None

# --------------------------------------------------------------------

comm = MPI.COMM_WORLD

buff = numpy.empty((1000, 2), dtype="f", order="f")

rank = comm.Get_rank()

if rank == 0:
    req1 = comm.Irecv([buff[:, 0], MPI.FLOAT], 1, 1)
    req2 = comm.Irecv([buff[:, 1], MPI.FLOAT], 1, 2)
    status = [MPI.Status(), MPI.Status()]
    MPI.Request.Waitall([req1, req2], status)
elif rank == 1:
    buff[:, 0] = 5
    buff[:, 1] = 7
    comm.Ssend([buff[:, 1], MPI.FLOAT], 0, 2)
    comm.Rsend([buff[:, 0], MPI.FLOAT], 0, 1)

# --------------------------------------------------------------------

if rank == 0:
    assert numpy.all(buff[:, 0] == 5)
    assert numpy.all(buff[:, 1] == 7)
    assert status[0].source == 1
    assert status[0].tag == 1
    assert status[1].source == 1
    assert status[1].tag == 2

# --------------------------------------------------------------------



import os
import pathlib
import tempfile

import mpiunittest as unittest

from mpi4py import MPI

# ---


class MyBaseComm:
    #
    def free(self):
        if self != MPI.COMM_NULL:
            MPI.Comm.Free(self)


class BaseTestBaseComm:
    #
    def setUp(self):
        self.comm = self.CommType(self.COMM_BASE)

    def testSubType(self):
        self.assertNotIn(
            type(self.comm),
            [
                MPI.Comm,
                MPI.Intracomm,
                MPI.Cartcomm,
                MPI.Graphcomm,
                MPI.Distgraphcomm,
                MPI.Intercomm,
            ],
        )
        self.assertIsInstance(self.comm, self.CommType)

    def testCloneFree(self):
        if self.COMM_BASE != MPI.COMM_NULL:
            comm = self.comm.Clone()
        else:
            comm = self.CommType()
        self.assertIsInstance(comm, MPI.Comm)
        self.assertIsInstance(comm, self.CommType)
        comm.free()

    def tearDown(self):
        self.comm.free()


# ---


class MyComm(MPI.Comm, MyBaseComm):
    #
    def __new__(cls, comm=None):
        if comm is not None:
            if comm != MPI.COMM_NULL:
                comm = comm.Clone()
        return super().__new__(cls, comm)


class BaseTestMyComm(BaseTestBaseComm):
    #
    CommType = MyComm


class TestMyCommNULL(BaseTestMyComm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_NULL


class TestMyCommSELF(BaseTestMyComm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_SELF


class TestMyCommWORLD(BaseTestMyComm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_WORLD


# ---


class MyIntracomm(MPI.Intracomm, MyBaseComm):
    #
    def __new__(cls, comm=None):
        if comm is not None:
            if comm != MPI.COMM_NULL:
                comm = comm.Dup()
        return super().__new__(cls, comm)


class BaseTestMyIntracomm(BaseTestBaseComm):
    #
    CommType = MyIntracomm


class TestMyIntracommNULL(BaseTestMyIntracomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_NULL


class TestMyIntracommSELF(BaseTestMyIntracomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_SELF


class TestMyIntracommWORLD(BaseTestMyIntracomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_WORLD


# ---


class MyCartcomm(MPI.Cartcomm, MyBaseComm):
    #
    def __new__(cls, comm=None):
        if comm is not None:
            if comm != MPI.COMM_NULL:
                dims = [comm.size]
                comm = comm.Create_cart(dims)
        return super().__new__(cls, comm)


class BaseTestMyCartcomm(BaseTestBaseComm):
    #
    CommType = MyCartcomm


class TestMyCartcommNULL(BaseTestMyCartcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_NULL


class TestMyCartcommSELF(BaseTestMyCartcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_SELF


class TestMyCartcommWORLD(BaseTestMyCartcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_WORLD


# ---


class MyGraphcomm(MPI.Graphcomm, MyBaseComm):
    #
    def __new__(cls, comm=None):
        if comm is not None:
            if comm != MPI.COMM_NULL:
                index = list(range(comm.size + 1))
                edges = list(range(comm.size))
                comm = comm.Create_graph(index, edges)
        return super().__new__(cls, comm)


class BaseTestMyGraphcomm(BaseTestBaseComm):
    #
    CommType = MyGraphcomm


class TestMyGraphcommNULL(BaseTestMyGraphcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_NULL


class TestMyGraphcommSELF(BaseTestMyGraphcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_SELF


class TestMyGraphcommWORLD(BaseTestMyGraphcomm, unittest.TestCase):
    #
    COMM_BASE = MPI.COMM_WORLD


# ---


class MyRequest(MPI.Request):
    #
    def __new__(cls, request=None):
        return super().__new__(cls, request)

    def test(self):
        return super(type(self), self).Test()

    def wait(self):
        return super(type(self), self).Wait()


class MyPrequest(MPI.Prequest):
    #
    def __new__(cls, request=None):
        return super().__new__(cls, request)

    def test(self):
        return super(type(self), self).Test()

    def wait(self):
        return super(type(self), self).Wait()

    def start(self):
        return super(type(self), self).Start()


class MyGrequest(MPI.Grequest):
    #
    def __new__(cls, request=None):
        return super().__new__(cls, request)

    def test(self):
        return super(type(self), self).Test()

    def wait(self):
        return super(type(self), self).Wait()


class BaseTestMyRequest:
    #
    def setUp(self):
        self.req = self.MyRequestType(MPI.REQUEST_NULL)

    def testSubType(self):
        self.assertIsNot(type(self.req), self.MPIRequestType)
        self.assertIsInstance(self.req, self.MPIRequestType)
        self.assertIsInstance(self.req, self.MyRequestType)
        self.req.test()


class TestMyRequest(BaseTestMyRequest, unittest.TestCase):
    #
    MPIRequestType = MPI.Request
    MyRequestType = MyRequest


class TestMyPrequest(BaseTestMyRequest, unittest.TestCase):
    #
    MPIRequestType = MPI.Prequest
    MyRequestType = MyPrequest


class TestMyGrequest(BaseTestMyRequest, unittest.TestCase):
    #
    MPIRequestType = MPI.Grequest
    MyRequestType = MyGrequest


class TestMyRequest2(TestMyRequest):
    #
    def setUp(self):
        req = MPI.COMM_SELF.Isend(
            [MPI.BOTTOM, 0, MPI.BYTE], dest=MPI.PROC_NULL, tag=0
        )
        self.req = MyRequest(req)


@unittest.skipMPI("mpich(==3.4.1)")
class TestMyPrequest2(TestMyPrequest):
    #
    def setUp(self):
        req = MPI.COMM_SELF.Send_init(
            [MPI.BOTTOM, 0, MPI.BYTE], dest=MPI.PROC_NULL, tag=0
        )
        self.req = MyPrequest(req)

    def tearDown(self):
        self.req.Free()

    def testStart(self):
        for _i in range(5):
            self.req.start()
            self.req.test()
            self.req.start()
            self.req.wait()


# ---


class MyInfo(MPI.Info):
    #
    def __new__(cls, info=None):
        return MPI.Info.__new__(cls, info)

    def free(self):
        if self != MPI.INFO_NULL:
            MPI.Info.Free(self)


class BaseTestMyInfo:
    #
    def setUp(self):
        info = MPI.Info.Create()
        self.info = MyInfo(info)

    def tearDown(self):
        self.info.free()

    def testSubType(self):
        self.assertIsNot(type(self.info), MPI.Info)
        self.assertIsInstance(self.info, MPI.Info)
        self.assertIsInstance(self.info, MyInfo)

    def testFree(self):
        self.assertTrue(self.info)
        self.info.free()
        self.assertFalse(self.info)

    def testCreateDupType(self):
        for info in (
            MyInfo.Create(),
            self.info.Dup(),
            self.info.copy(),
        ):
            self.assertIsNot(type(info), MPI.Info)
            self.assertIsInstance(info, MPI.Info)
            self.assertIsInstance(info, MyInfo)
            info.free()

    def testCreateEnvType(self):
        try:
            info = MyInfo.Create_env()
        except NotImplementedError:
            if MPI.Get_version() >= (4, 0):
                raise
            raise unittest.SkipTest("mpi-info-create-env") from None
        self.assertIsNot(type(info), MPI.Info)
        self.assertIsInstance(info, MPI.Info)
        self.assertIsInstance(info, MyInfo)

    def testPickle(self):
        from pickle import dumps, loads

        items = list(zip("abc", "123"))
        self.info.update(items)
        info = loads(dumps(self.info))
        self.assertIs(type(info), MyInfo)
        self.assertEqual(info.items(), items)
        info.free()


class TestMyInfo(BaseTestMyInfo, unittest.TestCase):
    #
    pass


try:
    MPI.Info.Create().Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestMyInfo, "mpi-info")

# ---


class MyWin(MPI.Win):
    #
    def __new__(cls, win=None):
        return MPI.Win.__new__(cls, win)

    def free(self):
        if self != MPI.WIN_NULL:
            MPI.Win.Free(self)


class BaseTestMyWin:
    #
    def setUp(self):
        w = MPI.Win.Create(MPI.BOTTOM)
        self.win = MyWin(w)

    def tearDown(self):
        self.win.free()

    def testSubType(self):
        self.assertIsNot(type(self.win), MPI.Win)
        self.assertIsInstance(self.win, MPI.Win)
        self.assertIsInstance(self.win, MyWin)

    def testFree(self):
        self.assertTrue(self.win)
        self.win.free()
        self.assertFalse(self.win)


class TestMyWin(BaseTestMyWin, unittest.TestCase):
    #
    pass


try:
    MPI.Win.Create(MPI.BOTTOM).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestMyWin, "mpi-win")

# ---


class MyFile(MPI.File):
    #
    def __new__(cls, file=None):
        return MPI.File.__new__(cls, file)

    def close(self):
        if self != MPI.FILE_NULL:
            MPI.File.Close(self)


class BaseTestMyFile:
    #
    def openfile(self):
        fd, fname = tempfile.mkstemp(prefix="mpi4py")
        os.close(fd)
        fname = pathlib.Path(fname)
        amode = MPI.MODE_RDWR | MPI.MODE_CREATE | MPI.MODE_DELETE_ON_CLOSE
        try:
            self.file = MPI.File.Open(
                MPI.COMM_SELF, fname, amode, MPI.INFO_NULL
            )
        except Exception:
            fname.unlink()
            raise
        return self.file

    def setUp(self):
        f = self.openfile()
        self.file = MyFile(f)

    def tearDown(self):
        self.file.close()

    def testSubType(self):
        self.assertIsNot(type(self.file), MPI.File)
        self.assertIsInstance(self.file, MPI.File)
        self.assertIsInstance(self.file, MyFile)

    def testFree(self):
        self.assertTrue(self.file)
        self.file.close()
        self.assertFalse(self.file)


class TestMyFile(BaseTestMyFile, unittest.TestCase):
    #
    pass


try:
    BaseTestMyFile().openfile().Close()
except NotImplementedError:
    unittest.disable(BaseTestMyFile, "mpi-file")

# ---

if __name__ == "__main__":
    unittest.main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Utilities for handling futures."""
# pylint: disable=too-few-public-methods
# pylint: disable=broad-exception-caught

import collections as _collections
import functools as _functools
import threading as _threading

from . import _base

__all__ = [
    "collect",
    "compose",
]


def collect(fs):
    """Gather a collection of futures in a new future.

    Args:
        fs: Collection of futures.

    Returns:
        New future producing as result a list with results from *fs*.

    """
    return _Collect()(fs)


class _Collect:
    #
    def __init__(self):
        self.lock = _threading.RLock()
        self.future = None
        self.result = None
        self.pending = None

    def __call__(self, fs):
        pending = _collections.defaultdict(list)
        for index, item in enumerate(fs):
            pending[item].append(index)
        if pending:
            self.future = future = next(iter(pending)).__class__()
            self.result = [None] * sum(map(len, pending.values()))
            self.pending = pending
            future.add_done_callback(self._done_cb)
            for item in list(pending):
                item.add_done_callback(self._item_cb)
        else:
            future = _base.Future()
            future.set_result([])
        return future

    def _item_cb(self, item):
        with self.lock:
            future = self.future
            result = self.result
            pending = self.pending
            if future is None:
                return
            if item.cancelled():
                future.cancel()
                return
            if item.exception() is not None:
                if future.set_running_or_notify_cancel():  # pragma: no branch
                    future.set_exception(item.exception())
                return
            for index in pending.pop(item):
                result[index] = item.result()
            if not pending:
                if future.set_running_or_notify_cancel():  # pragma: no branch
                    future.set_result(result)

    def _done_cb(self, future):
        with self.lock:
            for item in self.pending:
                item.cancel()
            self.future = None
            self.result = None
            self.pending = None
        if future.cancelled():
            future.set_running_or_notify_cancel()


def compose(future, resulthook=None, excepthook=None):
    """Compose the completion of a future with result and exception handlers.

    Args:
        future: Input future instance.
        resulthook: Function to be called once the input future completes with
            success. Once the input future finish running with success, its
            result value is the input argument for *resulthook*. The result of
            *resulthook* is set as the result of the output future.
            If *resulthook* is ``None``, the output future is completed
            directly with the result of the input future.
        excepthook: Function to be called once the input future completes with
            failure. Once the input future finish running with failure, its
            exception value is the input argument for *excepthook*.  If
            *excepthook* returns an `python:Exception` instance, it is set as
            the exception of the output future.  Otherwise, the result of
            *excepthook* is set as the result of the output future.  If
            *excepthook* is ``None``, the output future is set as failed with
            the exception from the input future.

    Returns:
        Output future instance to be completed once the input future is
        completed and either *resulthook* or *excepthook* finish executing.

    """
    new_future = future.__class__()
    new_done_cb = _functools.partial(_compose_cancel, [future])
    new_future.add_done_callback(new_done_cb)
    context = [(new_future, resulthook, excepthook)]
    done_cb = _functools.partial(_compose_complete, context)
    future.add_done_callback(done_cb)
    return new_future


def _compose_cancel(context, new_future):
    old_future = context.pop()
    try:
        if new_future.cancelled():
            new_future.set_running_or_notify_cancel()
            old_future.cancel()
    finally:
        new_future = None
        old_future = None


def _compose_complete(context, old_future):
    new_future, resulthook, excepthook = context.pop()
    try:
        if old_future.cancelled():
            new_future.cancel()
        elif old_future.exception() is not None:
            reason = old_future.exception()
            if excepthook:
                reason = excepthook(reason)
            if isinstance(reason, BaseException):
                new_future.set_exception(reason)
            else:
                new_future.set_result(reason)
        else:
            result = old_future.result()
            if resulthook:
                result = resulthook(result)
            new_future.set_result(result)
    except BaseException as exception:
        new_future.set_exception(exception)
    finally:
        new_future = None
        old_future = None
        result = resulthook = None
        reason = excepthook = None



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    ctypedef struct PyObject
    void Py_CLEAR(PyObject*)
    object PyLong_FromVoidPtr(void*)
    void*  PyLong_AsVoidPtr(object) except? NULL

# Python 3 buffer interface (PEP 3118)
cdef extern from "Python.h":
    ctypedef struct Py_buffer:
        PyObject *obj
        void *buf
        Py_ssize_t len
        Py_ssize_t itemsize
        bint readonly
        char *format
        int ndim
        Py_ssize_t *shape
        Py_ssize_t *strides
        Py_ssize_t *suboffsets
        void *internal
    cdef enum:
        PyBUF_SIMPLE
        PyBUF_WRITABLE
        PyBUF_FORMAT
        PyBUF_ND
        PyBUF_STRIDES
        PyBUF_ANY_CONTIGUOUS
    int  PyObject_CheckBuffer(object)
    int  PyObject_GetBuffer(object, Py_buffer *, int) except -1
    void PyBuffer_Release(Py_buffer *)
    int  PyBuffer_FillInfo(Py_buffer *, object,
                           void *, Py_ssize_t,
                           bint, int) except -1

cdef extern from "Python.h":
    enum: PyBUF_READ
    enum: PyBUF_WRITE
    object PyMemoryView_FromObject(object)
    object PyMemoryView_GetContiguous(object, int, char)


cdef inline int is_big_endian() noexcept nogil:
    cdef int i = 1
    return (<char*>&i)[0] == 0


cdef inline int is_little_endian() noexcept nogil:
    cdef int i = 1
    return (<char*>&i)[0] != 0


cdef char BYTE_FMT[2]
BYTE_FMT[0] = c'B'
BYTE_FMT[1] = 0

include "asdlpack.pxi"
include "ascaibuf.pxi"

cdef int PyMPI_GetBuffer(object obj, Py_buffer *view, int flags) except -1:
    try:
        return PyObject_GetBuffer(obj, view, flags)
    except BaseException:
        try:
            return Py_GetDLPackBuffer(obj, view, flags)
        except NotImplementedError:
            pass
        except BaseException:
            raise
        try:
            return Py_GetCAIBuffer(obj, view, flags)
        except NotImplementedError:
            pass
        except BaseException:
            raise
        raise

cdef void PyMPI_ReleaseBuffer(int kind, Py_buffer *view) noexcept:
    if kind == 0:
        # Python buffer interface
        PyBuffer_Release(view)
    else:
        # DLPack/CAI buffer interface
        Py_CLEAR(view.obj)

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int PyIndex_Check(object)
    int PySlice_Check(object)
    int PySlice_Unpack(object,
                       Py_ssize_t *,
                       Py_ssize_t *,
                       Py_ssize_t *) except -1
    Py_ssize_t PySlice_AdjustIndices(Py_ssize_t,
                                     Py_ssize_t *,
                                     Py_ssize_t *,
                                     Py_ssize_t) noexcept nogil
    Py_ssize_t PyNumber_AsSsize_t(object, object) except? -1

cdef inline int check_cpu_accessible(int kind) except -1:
    cdef unsigned device_type = <unsigned> kind
    if device_type == 0              : return 0
    if device_type == kDLCPU         : return 0
    if device_type == kDLCUDAHost    : return 0  # ~> uncovered
    if device_type == kDLROCMHost    : return 0  # ~> uncovered
    if device_type == kDLCUDAManaged : return 0  # ~> uncovered
    raise BufferError("buffer is not CPU-accessible")


@cython.final
cdef class buffer:
    """
    Buffer.
    """

    cdef Py_buffer view
    cdef int       kind

    def __cinit__(self, *args):
        if args:
            self.kind = PyMPI_GetBuffer(args[0], &self.view, PyBUF_SIMPLE)
        else:
            PyBuffer_FillInfo(&self.view, <object>NULL,
                              NULL, 0, 0, PyBUF_SIMPLE)

    def __dealloc__(self):
        PyMPI_ReleaseBuffer(self.kind, &self.view)

    @staticmethod
    def allocate(
        Aint nbytes: int,
        bint clear: bool = False,
    ) -> buffer:
        """Buffer allocation."""
        cdef void *addr = NULL
        cdef Py_ssize_t size = nbytes
        if size < 0:
            raise ValueError("expecting non-negative size")
        cdef object ob = rawalloc(size, 1, clear, &addr)
        cdef buffer buf = <buffer>New(buffer)
        PyBuffer_FillInfo(&buf.view, ob, addr, size, 0, PyBUF_SIMPLE)
        return buf

    @staticmethod
    def frombuffer(
        obj: Buffer,
        bint readonly: bool = False,
    ) -> buffer:
        """Buffer from buffer-like object."""
        cdef int flags = PyBUF_SIMPLE
        if not readonly: flags |= PyBUF_WRITABLE
        cdef buffer buf = <buffer>New(buffer)
        buf.kind = PyMPI_GetBuffer(obj, &buf.view, flags)
        buf.view.readonly = readonly
        return buf

    @staticmethod
    def fromaddress(
        address: int,
        Aint nbytes: int,
        bint readonly: bool = False,
    ) -> buffer:
        """Buffer from address and size in bytes."""
        cdef void *addr = PyLong_AsVoidPtr(address)
        cdef Py_ssize_t size = nbytes
        if size < 0:
            raise ValueError("expecting non-negative buffer length")
        elif size > 0 and addr == NULL:
            raise ValueError("expecting non-NULL address")
        cdef buffer buf = <buffer>New(buffer)
        PyBuffer_FillInfo(&buf.view, <object>NULL,
                          addr, size, readonly, PyBUF_SIMPLE)
        return buf

    # properties

    property address:
        """Buffer address."""
        def __get__(self) -> int:
            return PyLong_FromVoidPtr(self.view.buf)

    # memoryview properties

    property obj:
        """Object exposing buffer."""
        def __get__(self) -> Buffer | None:
            if self.view.obj == NULL: return None
            return <object>self.view.obj

    property nbytes:
        """Buffer size (in bytes)."""
        def __get__(self) -> int:
            return self.view.len

    property readonly:
        """Buffer is read-only."""
        def __get__(self) -> bool:
            return self.view.readonly

    property format:
        """Format of each element."""
        def __get__(self) -> str:
            if self.view.format != NULL:
                return pystr(self.view.format)
            return pystr(BYTE_FMT)

    property itemsize:
        """Size (in bytes) of each element."""
        def __get__(self) -> int:
            return self.view.itemsize

    # memoryview methods

    def cast(
        self,
        format: str,
        shape: list[int] | tuple[int, ...] = ...,
    ) -> memoryview:
        """
        Cast to a `memoryview` with new format or shape.
        """
        check_cpu_accessible(self.kind)
        if shape is Ellipsis:
            return PyMemoryView_FromObject(self).cast(format)
        else:
            return PyMemoryView_FromObject(self).cast(format, shape)

    def tobytes(self, order: str | None = None) -> bytes:
        """Return the data in the buffer as a byte string."""
        <void> order  # unused
        check_cpu_accessible(self.kind)
        return PyBytes_FromStringAndSize(<char*>self.view.buf, self.view.len)

    def toreadonly(self) -> buffer:
        """Return a readonly version of the buffer object."""
        cdef object obj = self
        if self.view.obj != NULL:
            obj = <object>self.view.obj
        cdef buffer buf = <buffer>New(buffer)
        buf.kind = PyMPI_GetBuffer(obj, &buf.view, PyBUF_SIMPLE)
        buf.view.readonly = 1
        return buf

    def release(self) -> None:
        """Release the underlying buffer exposed by the buffer object."""
        PyMPI_ReleaseBuffer(self.kind, &self.view)
        PyBuffer_FillInfo(&self.view, <object>NULL,
                          NULL, 0, 0, PyBUF_SIMPLE)
        self.kind = 0

    # buffer interface (PEP 3118)

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, self,
                          self.view.buf, self.view.len,
                          self.view.readonly, flags)

    # sequence interface (basic)

    def __len__(self):
        return self.view.len

    def __getitem__(self, object item):
        check_cpu_accessible(self.kind)
        cdef Py_ssize_t start=0, stop=0, step=1, slen=0
        cdef unsigned char *buf = <unsigned char*>self.view.buf
        cdef Py_ssize_t blen = self.view.len
        if PyIndex_Check(item):
            start = PyNumber_AsSsize_t(item, IndexError)
            if start < 0: start += blen
            if start < 0 or start >= blen:
                raise IndexError("index out of range")
            return <long>buf[start]
        elif PySlice_Check(item):
            PySlice_Unpack(item, &start, &stop, &step)
            slen = PySlice_AdjustIndices(blen, &start, &stop, step)
            if step != 1: raise IndexError("slice with step not supported")
            return tobuffer(self, buf+start, slen, self.view.readonly)
        else:
            raise TypeError("index must be integer or slice")

    def __setitem__(self, object item, object value):
        check_cpu_accessible(self.kind)
        if self.view.readonly:
            raise TypeError("buffer is read-only")
        cdef Py_ssize_t start=0, stop=0, step=1, slen=0
        cdef unsigned char *buf = <unsigned char*>self.view.buf
        cdef Py_ssize_t blen = self.view.len
        cdef buffer inbuf
        if PyIndex_Check(item):
            start = PyNumber_AsSsize_t(item, IndexError)
            if start < 0: start += blen
            if start < 0 or start >= blen:
                raise IndexError("index out of range")
            buf[start] = <unsigned char>value
        elif PySlice_Check(item):
            PySlice_Unpack(item, &start, &stop, &step)
            slen = PySlice_AdjustIndices(blen, &start, &stop, step)
            if step != 1: raise IndexError("slice with step not supported")
            if PyIndex_Check(value):
                <void>memset(buf+start, <unsigned char>value, <size_t>slen)
            else:
                inbuf = getbuffer(value, 1, 0)
                if inbuf.view.len != slen:
                    raise ValueError("slice length does not match buffer")
                <void>memmove(buf+start, inbuf.view.buf, <size_t>slen)
        else:
            raise TypeError("index must be integer or slice")


memory = buffer  # Backward compatibility alias

# -----------------------------------------------------------------------------

cdef inline buffer newbuffer():
    return <buffer>New(buffer)

cdef inline buffer getbuffer(object ob, bint readonly, bint format):
    cdef buffer buf = newbuffer()
    cdef int flags = PyBUF_ANY_CONTIGUOUS
    if not readonly:
        flags |= PyBUF_WRITABLE
    if format:
        flags |= PyBUF_FORMAT
    buf.kind = PyMPI_GetBuffer(ob, &buf.view, flags)
    return buf

cdef inline buffer asbuffer(object ob, void **base, MPI_Aint *size, bint ro):
    cdef buffer buf
    if type(ob) is buffer:
        buf = <buffer> ob
        if buf.view.readonly and not ro:
            raise BufferError("Object is not writable")
    else:
        buf = getbuffer(ob, ro, 0)
    if base != NULL: base[0] = buf.view.buf
    if size != NULL: size[0] = buf.view.len
    return buf

cdef inline buffer asbuffer_r(object ob, void **base, MPI_Aint *size):
    return asbuffer(ob, base, size, 1)

cdef inline buffer asbuffer_w(object ob, void **base, MPI_Aint *size):
    return asbuffer(ob, base, size, 0)

cdef inline buffer tobuffer(object ob, void *base, MPI_Aint size, bint ro):
    if size < 0: raise ValueError("expecting non-negative buffer length")
    cdef buffer buf = newbuffer()
    PyBuffer_FillInfo(&buf.view, ob, base, size, ro, PyBUF_SIMPLE)
    return buf

cdef inline buffer mpibuf(void *base, MPI_Count count):
    cdef MPI_Aint size = <MPI_Aint>count
    cdef int neq = (count != <MPI_Count>size)
    if neq: raise OverflowError("length {count} does not fit in 'MPI_Aint'")
    return tobuffer(<object>NULL, base, size, 0)

cdef inline object aspybuffer(
    object obj,
    void **base,
    MPI_Aint *size,
    bint readonly,
    const char format[],
):
    cdef int buftype = PyBUF_READ if readonly else PyBUF_WRITE
    obj = PyMemoryView_GetContiguous(obj, buftype, c'A')
    cdef Py_buffer view
    cdef int flags = PyBUF_ANY_CONTIGUOUS
    if not readonly:
        flags |= PyBUF_WRITABLE
    if format != NULL:
        flags |= PyBUF_FORMAT
    PyObject_GetBuffer(obj, &view, flags)
    if format != NULL and view.format != NULL:
        if strncmp(format, view.format, 4) != 0:
            PyBuffer_Release(&view)
            raise ValueError(
                f"expecting buffer with format {pystr(format)!r}, "
                f"got {pystr(view.format)!r}")
    if base != NULL: base[0] = view.buf
    if size != NULL: size[0] = view.len // view.itemsize
    PyBuffer_Release(&view)
    return obj

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef object datarep_lock     = Lock()
cdef dict   datarep_registry = {}


@cython.linetrace(False)  # ~> TODO
@cython.final
@cython.internal
cdef class _p_datarep:

    cdef object read_fn
    cdef object write_fn
    cdef object extent_fn

    def __cinit__(self, read_fn, write_fn, extent_fn):
        self.read_fn   = read_fn
        self.write_fn  = write_fn
        self.extent_fn = extent_fn

    cdef int read(
        self,
        void *userbuf,
        MPI_Datatype datatype,
        MPI_Count count,
        void *filebuf,
        MPI_Offset position,
    ) except -1:
        cdef MPI_Count lb=0, extent=0
        CHKERR( MPI_Type_get_extent_c(datatype, &lb, &extent) )
        cdef MPI_Count ulen = (position + count) * extent
        cdef MPI_Count flen = <MPI_Count> PY_SSIZE_T_MAX
        cdef object ubuf = mpibuf(userbuf, ulen)
        cdef object fbuf = mpibuf(filebuf, flen)
        cdef Datatype dtype = <Datatype>New(Datatype)
        dtype.ob_mpi = datatype
        try:
            self.read_fn(ubuf, dtype, count, fbuf, position)
        finally:
            dtype.ob_mpi = MPI_DATATYPE_NULL
        return 0

    cdef int write(
        self,
        void *userbuf,
        MPI_Datatype datatype,
        MPI_Count count,
        void *filebuf,
        MPI_Offset position,
    ) except -1:
        cdef MPI_Count lb=0, extent=0
        CHKERR( MPI_Type_get_extent_c(datatype, &lb, &extent) )
        cdef MPI_Count ulen = (position + count) * extent
        cdef MPI_Count flen = <MPI_Count> PY_SSIZE_T_MAX
        cdef object ubuf = mpibuf(userbuf, ulen)
        cdef object fbuf = mpibuf(filebuf, flen)
        cdef Datatype dtype = <Datatype>New(Datatype)
        dtype.ob_mpi = datatype
        try:
            self.write_fn(ubuf, dtype, count, fbuf, position)
        finally:
            dtype.ob_mpi = MPI_DATATYPE_NULL
        return 0

    cdef int extent(
        self,
        MPI_Datatype datatype,
        MPI_Aint *file_extent,
    ) except -1:
        cdef Datatype dtype = <Datatype>New(Datatype)
        dtype.ob_mpi = datatype
        try:
            file_extent[0] = self.extent_fn(dtype)
        finally:
            dtype.ob_mpi = MPI_DATATYPE_NULL
        return 0


# ---


@cython.linetrace(False)  # ~> TODO
cdef int datarep_read(
    void *userbuf,
    MPI_Datatype datatype,
    MPI_Count count,
    void *filebuf,
    MPI_Offset position,
    void *extra_state,
) noexcept with gil:
    cdef _p_datarep state = <_p_datarep>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.read(userbuf, datatype, count, filebuf, position)
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


@cython.linetrace(False)  # ~> TODO
cdef int datarep_write(
    void *userbuf,
    MPI_Datatype datatype,
    MPI_Count count,
    void *filebuf,
    MPI_Offset position,
    void *extra_state,
) noexcept with gil:
    cdef _p_datarep state = <_p_datarep>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.write(userbuf, datatype, count, filebuf, position)
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


@cython.linetrace(False)  # ~> TODO
cdef int datarep_extent(
    MPI_Datatype datatype,
    MPI_Aint *file_extent,
    void *extra_state,
) noexcept with gil:
    cdef _p_datarep state = <_p_datarep>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.extent(datatype, file_extent)
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


# ---


@cython.linetrace(False)  # ~> TODO
@cython.callspec("MPIAPI")
cdef int datarep_read_fn(
    void *userbuf,
    MPI_Datatype datatype,
    MPI_Count count,
    void *filebuf,
    MPI_Offset position,
    void *extra_state,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return datarep_read(
        userbuf, datatype, count,
        filebuf, position, extra_state,
    )


@cython.linetrace(False)  # ~> TODO
@cython.callspec("MPIAPI")
cdef int datarep_write_fn(
    void *userbuf,
    MPI_Datatype datatype,
    MPI_Count count,
    void *filebuf,
    MPI_Offset position,
    void *extra_state,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return datarep_write(
        userbuf, datatype, count,
        filebuf, position, extra_state,
    )


@cython.linetrace(False)  # ~> TODO
@cython.callspec("MPIAPI")
cdef int datarep_extent_fn(
    MPI_Datatype datatype,
    MPI_Aint *file_extent,
    void *extra_state,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return datarep_extent(
        datatype, file_extent, extra_state,
    )

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    void Py_INCREF(object)
    void Py_DECREF(object)

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_keyval:

    cdef public object copy_fn
    cdef public object delete_fn
    cdef public bint   nopython
    cdef int    ierr
    cdef object lock

    def __cinit__(self, copy_fn, delete_fn, nopython):
        if copy_fn   is False: copy_fn   = None
        if delete_fn is False: delete_fn = None
        if delete_fn is True:  delete_fn = None
        self.copy_fn   = copy_fn
        self.delete_fn = delete_fn
        self.nopython  = nopython
        self.ierr = MPI_SUCCESS
        self.lock = RLock()

cdef object keyval_lock_type = Lock()
cdef object keyval_lock_comm = Lock()
cdef object keyval_lock_win  = Lock()

cdef dict   keyval_registry_type = {}
cdef dict   keyval_registry_comm = {}
cdef dict   keyval_registry_win  = {}

_keyval_registry = {
    'Datatype' : keyval_registry_type,
    'Comm'     : keyval_registry_comm,
    'Win'      : keyval_registry_win,
}

# -----------------------------------------------------------------------------

ctypedef fused PyMPI_attr_type:
    MPI_Datatype
    MPI_Comm
    MPI_Win


cdef inline object PyMPI_attr_call(
    object function,
    PyMPI_attr_type hdl,
    int keyval,
    object attrval,
):
    cdef object handle
    cdef object result
    if PyMPI_attr_type is MPI_Datatype:
        handle = fromhandle(hdl)
    if PyMPI_attr_type is MPI_Comm:
        handle = fromhandle(hdl)
    if PyMPI_attr_type is MPI_Win:
        handle = fromhandle(hdl)
    try:
        result = function(handle, keyval, attrval)
    finally:
        if False: pass
        elif PyMPI_attr_type is MPI_Datatype:
            (<Datatype>handle).ob_mpi = MPI_DATATYPE_NULL
        elif PyMPI_attr_type is MPI_Comm:
            (<Comm>handle).ob_mpi = MPI_COMM_NULL
        elif PyMPI_attr_type is MPI_Win:
            (<Win>handle).ob_mpi = MPI_WIN_NULL
    return result


cdef inline int PyMPI_attr_copy(
    PyMPI_attr_type hdl,
    int keyval,
    void *extra_state,
    void *attrval_in,
    void *attrval_out,
    int *flag,
) except -1:
    if flag != NULL: flag[0] = 0
    cdef _p_keyval state = <_p_keyval>extra_state
    if state.copy_fn is None: return 0
    cdef int p = not state.nopython
    if p and attrval_in == NULL: raise RuntimeError
    cdef object attrval
    if p: attrval = <object>attrval_in
    else: attrval = PyLong_FromVoidPtr(attrval_in)
    if state.copy_fn is not True:
        attrval = PyMPI_attr_call(state.copy_fn, hdl, keyval, attrval)
        if attrval is NotImplemented: return 0
    cdef void **outval = <void **>attrval_out
    if p: outval[0] = <void *>attrval
    else: outval[0] = PyLong_AsVoidPtr(attrval)
    if flag != NULL: flag[0] = 1
    if p: Py_INCREF(attrval)
    Py_INCREF(state)
    return 0


cdef inline int PyMPI_attr_delete(
    PyMPI_attr_type hdl,
    int keyval,
    void *attrval_in,
    void *extra_state,
) except -1:
    cdef _p_keyval state = <_p_keyval>extra_state
    cdef int p = not state.nopython
    if p and attrval_in == NULL: raise RuntimeError
    cdef object attrval
    if p: attrval = <object>attrval_in
    else: attrval = PyLong_FromVoidPtr(attrval_in)
    try:
        if state.delete_fn is not None:
            PyMPI_attr_call(state.delete_fn, hdl, keyval, attrval)
    finally:
        if p: Py_DECREF(attrval)
        Py_DECREF(state)
    return 0


cdef inline int PyMPI_attr_copy_cb(
    PyMPI_attr_type hdl,
    int keyval,
    void *extra_state,
    void *attrval_in,
    void *attrval_out,
    int *flag,
) noexcept with gil:
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        PyMPI_attr_copy(
            hdl, keyval, extra_state,
            attrval_in, attrval_out, flag,
        )
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


cdef inline int PyMPI_attr_delete_cb(
    PyMPI_attr_type hdl,
    int keyval,
    void *attrval,
    void *extra_state,
) noexcept with gil:
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    cdef _p_keyval state = <_p_keyval>extra_state
    state.ierr = MPI_SUCCESS
    try:
        PyMPI_attr_delete(
            hdl, keyval, attrval, extra_state,
        )
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
        state.ierr, ierr = ierr, MPI_SUCCESS
    return ierr


@cython.callspec("MPIAPI")
cdef int PyMPI_attr_copy_fn(
    PyMPI_attr_type hdl,
    int keyval,
    void *extra_state,
    void *attrval_in,
    void *attrval_out,
    int *flag,
) noexcept nogil:
    if flag != NULL: flag[0] = 0
    if extra_state == NULL:    return MPI_ERR_INTERN
    if attrval_out == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_SUCCESS
    if not py_module_alive():  return MPI_SUCCESS
    return PyMPI_attr_copy_cb(
        hdl, keyval, extra_state,
        attrval_in, attrval_out, flag,
    )


@cython.callspec("MPIAPI")
cdef int PyMPI_attr_delete_fn(
    PyMPI_attr_type hdl,
    int keyval,
    void *attrval,
    void *extra_state,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_SUCCESS
    if not py_module_alive():  return MPI_SUCCESS
    return PyMPI_attr_delete_cb(
        hdl, keyval, attrval, extra_state,
    )

# ---

cdef inline _p_keyval PyMPI_attr_state_get(
    PyMPI_attr_type hdl,
    int keyval,
):
    <void> hdl  # unused
    if PyMPI_attr_type is MPI_Datatype:
        with keyval_lock_type:
            return <_p_keyval>keyval_registry_type.get(keyval)
    if PyMPI_attr_type is MPI_Comm:
        with keyval_lock_comm:
            return <_p_keyval>keyval_registry_comm.get(keyval)
    if PyMPI_attr_type is MPI_Win:
        with keyval_lock_win:
            return <_p_keyval>keyval_registry_win.get(keyval)


cdef inline int PyMPI_attr_state_set(
    PyMPI_attr_type hdl,
    int keyval,
    _p_keyval state,
) except -1:
    <void> hdl  # unused
    if PyMPI_attr_type is MPI_Datatype:
        with keyval_lock_type:
            keyval_registry_type[keyval] = state
    if PyMPI_attr_type is MPI_Comm:
        with keyval_lock_comm:
            keyval_registry_comm[keyval] = state
    if PyMPI_attr_type is MPI_Win:
        with keyval_lock_win:
            keyval_registry_win[keyval] = state
    return 0


cdef inline int PyMPI_attr_state_del(
    PyMPI_attr_type hdl,
    int keyval,
) except -1:
    <void> hdl  # unused
    try:
        if PyMPI_attr_type is MPI_Datatype:
            with keyval_lock_type:
                del keyval_registry_type[keyval]
        if PyMPI_attr_type is MPI_Comm:
            with keyval_lock_comm:
                del keyval_registry_comm[keyval]
        if PyMPI_attr_type is MPI_Win:
            with keyval_lock_win:
                del keyval_registry_win[keyval]
    except KeyError:  # ~> uncovered
        pass
    return 0

# ---

cdef inline object PyMPI_attr_get(
    PyMPI_attr_type hdl,
    int keyval,
    void *attrval,
):
    cdef _p_keyval state = PyMPI_attr_state_get(hdl, keyval)
    if state is not None and not state.nopython:
        return <object>attrval
    else:
        return PyLong_FromVoidPtr(attrval)


cdef inline int PyMPI_set_attr(
    PyMPI_attr_type hdl,
    int keyval,
    void *attrval,
) except -1:
    if PyMPI_attr_type is MPI_Datatype:
        CHKERR( MPI_Type_set_attr(hdl, keyval, attrval) )
    if PyMPI_attr_type is MPI_Comm:
        CHKERR( MPI_Comm_set_attr(hdl, keyval, attrval) )
    if PyMPI_attr_type is MPI_Win:
        CHKERR( MPI_Win_set_attr(hdl, keyval, attrval) )
    return 0

cdef inline int PyMPI_attr_set(
    PyMPI_attr_type hdl,
    int keyval,
    object attrval,
) except -1:
    cdef _p_keyval state = PyMPI_attr_state_get(hdl, keyval)
    cdef void *valptr = NULL
    if state is not None and not state.nopython:
        valptr = <void *>attrval
    else:
        valptr = PyLong_AsVoidPtr(attrval)
    PyMPI_set_attr(hdl, keyval, valptr)
    if state is not None:
        if not state.nopython:
            Py_INCREF(attrval)
        Py_INCREF(state)
    return 0

cdef inline int PyMPI_delete_attr(
    PyMPI_attr_type hdl,
    int keyval,
) except -1:
    if PyMPI_attr_type is MPI_Datatype:
        CHKERR( MPI_Type_delete_attr(hdl, keyval) )
    if PyMPI_attr_type is MPI_Comm:
        CHKERR( MPI_Comm_delete_attr(hdl, keyval) )
    if PyMPI_attr_type is MPI_Win:
        CHKERR( MPI_Win_delete_attr(hdl, keyval) )
    return 0

cdef inline int PyMPI_attr_del(
    PyMPI_attr_type hdl,
    int keyval,
) except -1:
    cdef int ierr = MPI_SUCCESS
    cdef _p_keyval state = PyMPI_attr_state_get(hdl, keyval)
    if state is None: return PyMPI_delete_attr(hdl, keyval)
    with state.lock:
        PyMPI_delete_attr(hdl, keyval)
        ierr, state.ierr = state.ierr, ierr
        CHKERR( ierr )

# -----------------------------------------------------------------------------
